# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

PositionRank keyphrase extraction model.

PositionRank is an unsupervised model for keyphrase extraction from scholarly
documents that incorporates information from all positions of a word's
occurrences into a biased PageRank. The model is described in:

* Corina Florescu and Cornelia Caragea.
  PositionRank: An Unsupervised Approach to Keyphrase Extraction from Scholarly
  Documents.
  *In proceedings of ACL*, pages 1105-1115, 2017.
"""

from collections import defaultdict
import networkx as nx

from pke_zh.unsupervised.singlerank import SingleRank


class PositionRank(SingleRank):
    """PositionRank keyphrase extraction model"""

    def __init__(self):
        """Redefining initializer for PositionRank."""

        super(PositionRank, self).__init__()

        self.positions = defaultdict(float)
        """Container the sums of word's inverse positions."""

    def clear_cache(self):
        # clear previous cache
        self.positions.clear()
        self.graph.clear()

    def candidate_selection(
            self,
            pos=None,
            maximum_word_number=3,
            **kwargs
    ):
        """Candidate selection heuristic using a syntactic PoS pattern for noun phrase extraction.

        :param pos: valid postag set
        :param maximum_word_number:the maximum number of words allowed for
                keyphrase candidates, defaults to 3.
        :param kwargs:
        :return:
        """
        pos = {'n', 'a'}
        self.longest_pos_sequence_selection(valid_pos=pos)

        # filter candidates greater than 3 words
        for k in list(self.candidates):
            v = self.candidates[k]
            if len(v.lexical_form) > maximum_word_number:
                del self.candidates[k]

    @staticmethod
    def is_valid_edge(idx2, text_length, position1, text, window):
        return idx2 < text_length and (text[idx2][1] - position1) < window

    def add_graph_edges(self, text, window):
        for i, (node1, position1) in enumerate(text):
            j = i + 1
            while self.is_valid_edge(j, len(text), position1, text, window):
                node2, _ = text[j]
                if node1 != node2:
                    if not self.graph.has_edge(node1, node2):
                        self.graph.add_edge(node1, node2, weight=0)
                    self.graph[node1][node2]['weight'] += 1
                j = j + 1

    def compute_word_reverse_position(self, text):
        for word, position in text:
            self.positions[word] += 1 / (position + 1)

    @staticmethod
    def filter_word_pos(sentence, pos, shift):
        valid_text = []
        for j, word in enumerate(sentence.words):
            if sentence.pos[j] in pos:
                valid_text.append((word, shift + j))
        return valid_text

    def build_word_graph(self, window=10, pos=None):
        """Build the graph representation of the document.

        In the graph, nodes are words that passes a Part-of-Speech filter. Two
        nodes are connected if the words corresponding to these nodes co-occur
        within a `window` of contiguous tokens. The weight of an edge is
        computed based on the co-occurrence count of the two words within a
        `window` of successive tokens.

        :param window: int, the window within the sentence for connecting two
            words in the graph, defaults to 10.
        :param pos: set, the set of valid pos for words to be considered as nodes
            in the graph, defaults to ('NOUN', 'PROP', 'ADJ').
        """
        if pos is None:
            pos = {'n', 'a'}

        # flatten document as a sequence of only valid (word, position) tuples
        text = []
        for i, sentence in enumerate(self.sentences):
            shift = sum([s.length for s in self.sentences[0:i]])
            valid_word_shifts = self.filter_word_pos(sentence, pos, shift)
            text.extend(valid_word_shifts)
        # add nodes to the graph
        self.graph.add_nodes_from([word for (word, position) in text])

        # add edges to the graph
        self.add_graph_edges(text, window)

        # compute the sums of the word's inverse positions
        self.compute_word_reverse_position(text)

    def candidate_weight_norm(self, pagerank_weight, normalized):
        for k in self.candidates.keys():
            tokens = self.candidates[k].lexical_form
            self.weights[k] = sum([pagerank_weight.get(t, 0.0) for t in tokens])
            if normalized:
                self.weights[k] /= len(tokens)

    def candidate_weighting(self, window=10, pos=None, normalized=False):
        """Candidate weight calculation using a biased PageRank.

        :param window: int, the window within the sentence for connecting two
                words in the graph, defaults to 10.
        :param pos: the set of valid pos for words to be considered as nodes
                in the graph, defaults to ('NOUN', 'PROPN', 'ADJ').
        :param normalized: bool, normalize keyphrase score by their length,
                defaults to False.
        """
        if pos is None:
            pos = {'n', 'a'}

        # build the word graph
        self.build_word_graph(window=window, pos=pos)
        # normalize cumulated inverse positions
        norm = sum(self.positions.values())
        for word in self.positions:
            self.positions[word] /= norm

        # compute the word scores using biased random walk
        w = nx.pagerank(G=self.graph,
                        alpha=0.85,
                        tol=0.0001,
                        personalization=self.positions,
                        weight='weight')

        # loop through the candidates
        self.candidate_weight_norm(w, normalized)

    def extract(self, input_file_or_string, n_best=10, pos=None):
        # 1. valid postags
        if pos is None:
            pos = {'n', 'a'}
        # 2. clear cache
        self.clear_cache()
        # 3. load the content of the document.
        self.load_document(input=input_file_or_string, language='zh', normalization=None)
        # 4. select the noun phrases up to 3 words as keyphrase candidates.
        self.candidate_selection(pos=pos, maximum_word_number=3)
        self.candidate_weighting(window=10, pos=pos)
        keyphrases = self.get_n_best(n=n_best)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

TopicRank keyphrase extraction model.

Graph-based ranking approach to keyphrase extraction described in:

* Adrien Bougouin, Florian Boudin and Béatrice Daille.
  TopicRank: Graph-Based Topic Ranking for Keyphrase Extraction.
  *In proceedings of IJCNLP*, pages 543-551, 2013.

"""

from itertools import combinations

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist

from pke_zh.base import BaseKeywordExtractModel


class TopicRank(BaseKeywordExtractModel):
    """TopicRank keywords extraction model."""

    def __init__(self, stopwords_path=None):
        """Redefining initializer for TopicRank."""

        super(TopicRank, self).__init__(stopwords_path)

        self.graph = nx.Graph()
        """ The topic graph. """

        self.topics = []
        """ The topic container. """

    def clear_cache(self):
        """clear previous graph and topics if exists"""
        self.graph.clear()
        self.topics.clear()

    def candidate_selection(self, pos=None, stoplist=None):
        """Selects longest sequences of nouns and adjectives as keyphrase candidates.

        :param pos: the set of valid POS tags, defaults to ('NOUN', 'ADJ').
        :param stoplist: the stoplist for filtering candidates, defaults to
            the nltk stoplist. Words that are punctuation marks from
            string.punctuation are not allowed.
        """
        # define default pos tags set
        if pos is None:
            pos = {'n', 'a'}

        # select sequence of adjectives and nouns
        self.longest_pos_sequence_selection(valid_pos=pos)

        # initialize stoplist list if not provided
        if stoplist is None:
            stoplist = self.stoplist

        # filter candidates containing stopwords or punctuation marks
        self.candidate_filtering(stoplist=stoplist)

    def vectorize_candidates(self):
        """Vectorize the keyphrase candidates.

        Returns:
            candidate_list (list): the list of candidates.
            candidate_matrix (matrix): vectorized representation of the candidates.
        """
        # build the vocabulary, i.e. setting the vector dimensions
        dim = set([])
        # for k, v in self.candidates.iteritems():
        # iterate Python 2/3 compatible
        for (k, v) in self.candidates.items():
            for w in v.lexical_form:
                dim.add(w)
        dim = list(dim)

        # vectorize the candidates Python 2/3 + sort for random issues
        candidate_list = list(self.candidates)
        candidate_list.sort()

        candidate_matrix = np.zeros((len(candidate_list), len(dim)))
        for i, k in enumerate(candidate_list):
            for w in self.candidates[k].lexical_form:
                candidate_matrix[i, dim.index(w)] += 1

        return candidate_list, candidate_matrix

    def topic_clustering(self, threshold=0.74, method='average'):
        """Clustering candidates into topics.

        :param threshold: float, the minimum similarity for clustering, defaults
            to 0.74, i.e. more than 1/4 of stem overlap similarity.
        :param method: str, the linkage method, defaults to average.
        """
        # handle document with only one candidate
        if len(self.candidates) == 1:
            self.topics.append([list(self.candidates)[0]])
            return

        # vectorize the candidates
        candidates, distance_matrix = self.vectorize_candidates()

        # compute the distance matrix
        jaccard_matrix = pdist(distance_matrix, 'jaccard')

        # compute the clusters
        cluster_matrix = linkage(jaccard_matrix, method=method)

        # form flat clusters
        clusters = fcluster(cluster_matrix, t=threshold, criterion='distance')

        # for each topic identifier
        for cluster_id in range(1, max(clusters) + 1):
            self.topics.append([candidates[j] for j in range(len(clusters))
                                if clusters[j] == cluster_id])

    @staticmethod
    def fetch_topic_index(topic_i, topic_j, offset_i, offset_j):
        if offset_i < offset_j:
            return topic_i
        else:
            return topic_j

    def calculate_topics_weight(self, ti, tj):
        """
        Calculate the weight of the topic.
        :param ti: topic_idx_i
        :param tj: topic_idx_j
        :return:
        """
        for c_i in self.topics[ti]:
            for c_j in self.topics[tj]:
                for p_i in self.candidates[c_i].offsets:
                    for p_j in self.candidates[c_j].offsets:
                        gap = abs(p_i - p_j)
                        target_c = self.fetch_topic_index(c_i, c_j, p_i, p_j)
                        gap -= len(self.candidates[target_c].lexical_form) - 1
                        self.graph[ti][tj]['weight'] += 1.0 / gap

    def build_topic_graph(self):
        """Build topic graph."""
        # adding the nodes to the graph
        self.graph.add_nodes_from(range(len(self.topics)))

        # loop through the topics to connect the nodes
        for i, j in combinations(range(len(self.topics)), 2):
            self.graph.add_edge(i, j, weight=0.0)
            self.calculate_topics_weight(i, j)

    def candidate_weighting(
            self,
            threshold=0.74,
            method='average',
            heuristic=None
    ):
        """Candidate ranking using random walk.

        :param threshold: float, the minimum similarity for clustering, defaults to 0.74.
        :param method: str, the linkage method, defaults to average.
        :param heuristic: str, the heuristic for selecting the best candidate for
            each topic, defaults to first occurring candidate. Other options
            are 'frequent' (most frequent candidate, position is used for
            ties).
        """
        if not self.candidates:
            return

        # cluster the candidates
        self.topic_clustering(threshold=threshold, method=method)

        # build the topic graph
        self.build_topic_graph()

        # compute the word scores using random walk
        w = nx.pagerank_scipy(self.graph, alpha=0.85, weight='weight')

        # loop through the topics
        for i, topic in enumerate(self.topics):

            # get the offsets of the topic candidates
            offsets = [self.candidates[t].offsets[0] for t in topic]

            # get first candidate from topic
            if heuristic == 'frequent':

                # get frequencies for each candidate within the topic
                freq = [len(self.candidates[t].surface_forms) for t in topic]

                # get the indexes of the most frequent candidates
                indexes = [j for j, f in enumerate(freq) if f == max(freq)]

                # offsets of the indexes
                indexes_offsets = [offsets[j] for j in indexes]
                # Choosing the first occuring most frequent candidate
                most_frequent = offsets.index(min(indexes_offsets))
                self.weights[topic[most_frequent]] = w[i]

            else:
                first = offsets.index(min(offsets))
                self.weights[topic[first]] = w[i]

    def extract(self, input_file_or_string, n_best=10, pos=None):
        if pos is None:
            pos = {'n', 'a'}
        self.load_document(input=input_file_or_string, language='zh')
        self.clear_cache()
        self.candidate_selection(pos=pos)
        self.candidate_weighting(threshold=0.74, method='average')
        # get the 10-highest scored candidates as keyphrases
        keyphrases = self.get_n_best(n=n_best)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin, Seon
@description:

YAKE(Yet Another Keyword Extractor)，一种基于关键词统计的单文档无监督关键词提取算法：
基于5种指标：是否大写，词的位置，词频，上下文关系，词在句中频率，来计算候选词的得分，从而筛选Top-N关键词。
中文只用后4个指标。
关键词的基本过滤规则：1）候选短语限制最长三个词；2）候选短语不能是专名；3）候选短语开头和结尾不能是停用词

YAKE keyphrase extraction model.

Statistical approach to keyphrase extraction described in:

* Ricardo Campos, Vítor Mangaravite, Arian Pasquali, Alípio Mário Jorge,
  Célia Nunes and Adam Jatowt.
  YAKE! Keyword extraction from single documents using multiple local features.
  *Information Sciences*, pages 257-289, 2020.

modify from: https://pypi.org/project/iyake-cn/0.5.5/#files

"""

import math
from collections import defaultdict
import numpy as np
from pke_zh.base import BaseKeywordExtractModel


class Yake(BaseKeywordExtractModel):
    """YAKE keyphrase extraction model."""

    def __init__(self):
        """Redefining initializer for YAKE.
        """
        super(Yake, self).__init__()
        self.words = defaultdict(set)
        """ Container for the vocabulary. """

        self.contexts = defaultdict(lambda: ([], []))
        """ Container for word contexts. """

        self.features = defaultdict(dict)
        """ Container for word features. """

        self.surface_to_lexical = {}
        """ Mapping from surface form to lexical form. """

    def candidate_selection(self, n=3):
        """Select 1-3 grams as keyphrase candidates. Candidates beginning or
        ending with a stopword are filtered out. Words that do not contain
        at least one alpha-numeric character are not allowed.

        Args:
            n (int): the n-gram length, defaults to 3.
        """
        # select ngrams from 1 to 3 grams
        self.ngram_selection(n=n)

        # filter candidates containing punctuation marks and stopwords
        self.candidate_filtering(stoplist=self.stoplist)

        # further filter candidates
        for k in list(self.candidates):
            # get the candidate
            v = self.candidates[k]

            # filter candidates starting/ending with a stopword
            if v.surface_forms[0][0].lower() in self.stoplist or v.surface_forms[0][
                -1].lower() in self.stoplist:
                del self.candidates[k]

    def _vocabulary_building(self):
        """Build the vocabulary that will be used to weight candidates. Only
        words containing at least one alpha-numeric character are kept.
        """
        # loop through sentences
        for i, sentence in enumerate(self.sentences):
            # compute the offset shift for the sentence
            shift = sum([s.length for s in self.sentences[0:i]])
            # loop through words in sentence
            for j, word in enumerate(sentence.words):
                # consider words containing at least one character
                if len(word) > 1:
                    # get the word
                    index = word.lower()
                    # add the word occurrence
                    self.words[index].add((shift + j, shift, i, word))

    def _contexts_building(self, window=2):
        """Build the contexts of the words for computing the relatedness
        feature. Words that occur within a window of n words are considered as
        context words. Only words co-occurring in a block (sequence of words
        that appear in the vocabulary) are considered.

        Args:
            use_stems (bool): whether to use stems instead of lowercase words
                for weighting, defaults to False.
            window (int): the size in words of the window used for computing
                co-occurrence counts, defaults to 2.
        """
        # loop through sentences
        for i, sentence in enumerate(self.sentences):
            # lowercase the words
            words = [w.lower() for w in sentence.words]
            # block container
            block = []
            # loop through words in sentence
            for j, word in enumerate(words):
                # skip and flush block if word is not in vocabulary
                if word not in self.words:
                    block = []
                    continue
                # add the left context
                self.contexts[word][0].extend(
                    [w for w in block[max(0, len(block) - window):len(block)]]
                )
                # add the right context
                for w in block[max(0, len(block) - window):len(block)]:
                    self.contexts[w][1].append(word)
                # add word to the current block
                block.append(word)

    def _feature_extraction(self):
        """Compute the weight of individual words using the following five
        features:

            1. CASING: gives importance to acronyms or words starting with a
               capital letter.

               CASING(w) = max(TF(U(w)), TF(A(w))) / (1 + log(TF(w)))

               with TF(U(w) being the # times the word starts with an uppercase
               letter, excepts beginning of sentences. TF(A(w)) is the # times
               the word is marked as an acronym.

            2. POSITION: gives importance to words occurring at the beginning of
               the document.

               POSITION(w) = log( log( 3 + Median(Sen(w)) ) )

               with Sen(w) contains the position of the sentences where w
               occurs.

            3. FREQUENCY: gives importance to frequent words.

               FREQUENCY(w) = TF(w) / ( MEAN_TF + STD_TF)

               with MEAN_TF and STD_TF computed on valid_tfs which are words
               that are not stopwords.

            4. RELATEDNESS: gives importance to words that do not have the
               characteristics of stopwords.

               RELATEDNESS(w) = 1 + (WR+WL)*(TF(w)/MAX_TF) + PL + PR

            5. DIFFERENT: gives importance to words that occurs in multiple
               sentences.

               DIFFERENT(w) = SF(w) / # sentences

               with SF(w) being the sentence frequency of word w.
        """
        # get the Term Frequency of each word
        TF = [len(self.words[w]) for w in self.words]

        # get the Term Frequency of non-stop words
        TF_nsw = [len(self.words[w]) for w in self.words if w not in self.stoplist]

        # compute statistics
        mean_TF = np.mean(TF_nsw)
        std_TF = np.std(TF_nsw)
        max_TF = max(TF)

        # Loop through the words
        for word in self.words:
            # Indicating whether the word is a stopwords
            self.features[word]['isstop'] = word in self.stoplist

            # Term Frequency
            self.features[word]['TF'] = len(self.words[word])

            # Uppercase/Acronym Term Frequencies
            self.features[word]['TF_A'] = 0
            self.features[word]['TF_U'] = 0
            for (offset, shift, sent_id, surface_form) in self.words[word]:
                if surface_form.isupper() and len(word) > 1:
                    self.features[word]['TF_A'] += 1
                elif surface_form[0].isupper() and offset != shift:
                    self.features[word]['TF_U'] += 1

            # 1. CASING feature, chinese text continue
            self.features[word]['CASING'] = max(self.features[word]['TF_A'],
                                                self.features[word]['TF_U'])
            self.features[word]['CASING'] /= 1.0 + math.log(self.features[word]['TF'])

            # 2. POSITION feature
            sentence_ids = list(set([t[2] for t in self.words[word]]))
            self.features[word]['POSITION'] = math.log(3.0 + np.median(sentence_ids))
            self.features[word]['POSITION'] = math.log(self.features[word]['POSITION'])

            # 3. FREQUENCY feature
            self.features[word]['FREQUENCY'] = self.features[word]['TF']
            self.features[word]['FREQUENCY'] /= (mean_TF + std_TF)

            # 4. RELATEDNESS feature
            self.features[word]['WL'] = 0.0
            if len(self.contexts[word][0]):
                self.features[word]['WL'] = len(set(self.contexts[word][0]))
                self.features[word]['WL'] /= len(self.contexts[word][0])
            self.features[word]['PL'] = len(set(self.contexts[word][0])) / max_TF

            self.features[word]['WR'] = 0.0
            if len(self.contexts[word][1]):
                self.features[word]['WR'] = len(set(self.contexts[word][1]))
                self.features[word]['WR'] /= len(self.contexts[word][1])
            self.features[word]['PR'] = len(set(self.contexts[word][1])) / max_TF

            self.features[word]['RELATEDNESS'] = 1
            self.features[word]['RELATEDNESS'] += (self.features[word]['WR'] +
                                                   self.features[word]['WL']) * \
                                                  (self.features[word]['TF'] / max_TF)

            # 5. DIFFERENT feature
            self.features[word]['DIFFERENT'] = len(set(sentence_ids))
            self.features[word]['DIFFERENT'] /= len(self.sentences)

            # assemble the features to weight words
            A = self.features[word]['CASING']
            B = self.features[word]['POSITION']
            C = self.features[word]['FREQUENCY']
            D = self.features[word]['RELATEDNESS']
            E = self.features[word]['DIFFERENT']
            self.features[word]['weight'] = (D * B) / (A + (C / D) + (E / D))

    def clear_cache(self):
        # clear previous cache
        self.words = defaultdict(set)
        self.contexts = defaultdict(lambda: ([], []))
        self.features = defaultdict(dict)
        self.surface_to_lexical = {}

    def candidate_weighting(self, window=2):
        """Candidate weight calculation as described in the YAKE paper.

        Args:
            window (int): the size in words of the window used for computing
                co-occurrence counts, defaults to 2.
        """
        if not self.candidates:
            return

        # build the vocabulary
        self._vocabulary_building()

        # extract the contexts
        self._contexts_building(window=window)

        # compute the word features
        self._feature_extraction()

        # compute candidate weights
        for k, v in self.candidates.items():
            lowercase_forms = [' '.join(t).lower() for t in v.surface_forms]
            for i, candidate in enumerate(lowercase_forms):
                TF = lowercase_forms.count(candidate)

                # computing differentiated weights for words and stopwords
                tokens = [t.lower() for t in v.surface_forms[i]]
                prod_ = 1.
                sum_ = 0.
                for j, token in enumerate(tokens):
                    if self.features[token]['isstop']:
                        term_stop = token
                        prob_t1 = 0
                        prob_t2 = 0
                        if j - 1 >= 0:
                            term_left = tokens[j - 1]
                            prob_t1 = self.contexts[term_left][1].count(
                                term_stop) / self.features[term_left]['TF']
                        if j + 1 < len(tokens):
                            term_right = tokens[j + 1]
                            prob_t2 = self.contexts[term_stop][0].count(
                                term_right) / self.features[term_right]['TF']

                        prob = prob_t1 * prob_t2
                        prod_ *= (1 + (1 - prob))
                        sum_ -= (1 - prob)
                    else:
                        prod_ *= self.features[token]['weight']
                        sum_ += self.features[token]['weight']
                if sum_ == -1:
                    # The candidate is a one token stopword at the start or
                    #  the end of the sentence
                    # Setting sum_ to -1+eps so 1+sum_ != 0
                    sum_ = -0.99999999999
                self.weights[candidate] = prod_
                self.weights[candidate] /= TF * (1 + sum_)
                self.surface_to_lexical[candidate] = k

    def is_redundant(self, candidate, prev, threshold=0.6):
        """Test if one candidate is redundant with respect to a list of already
        selected candidates. A candidate is considered redundant if its
        levenshtein distance, with another candidate that is ranked higher in
        the list, is greater than a threshold.

        Args:
            candidate (str): the lexical form of the candidate.
            prev (list): the list of already selected candidates.
            threshold (float): the threshold used when computing the
                    char sim score, defaults to 0.6.
        """
        # loop through the already selected candidates
        for prev_candidate in prev:
            s1 = set(prev_candidate)
            s2 = set(candidate)
            sim_score = len(s1 & s2) / min(len(s1), len(s2))
            if sim_score >= threshold:
                return True
        return False

    def get_n_best(
            self,
            n=10,
            redundancy_removal=True,
            threshold=0.6
    ):
        """ Returns the n-best candidates given the weights.

            Args:
                n (int): the number of candidates, defaults to 10.
                redundancy_removal (bool): whether redundant keyphrases are
                    filtered out from the n-best list using levenshtein
                    distance, defaults to True.
                threshold (float): the threshold used when computing the
                    char sim score, defaults to 0.6.
        """
        # sort candidates by ascending weight, change dict to key list
        best = sorted(self.weights, key=self.weights.get, reverse=True)

        # remove redundant candidates
        if redundancy_removal:
            # initialize a new container for non redundant candidates
            non_redundant_best = []
            # loop through the best candidates
            for candidate in best:
                # test weather candidate is redundant
                if self.is_redundant(candidate,
                                     non_redundant_best,
                                     threshold=threshold):
                    continue
                # add the candidate otherwise
                non_redundant_best.append(candidate)
                # break computation if the n-best are found
                if len(non_redundant_best) >= n:
                    break
            # copy non redundant candidates in best container
            best = non_redundant_best

        # get the list of best candidates as (lexical form, weight) tuples
        n_best = [(u.replace(' ', ''), self.weights[u]) for u in best[:min(n, len(best))]]
        # return the list of best candidates
        return n_best

    def extract(self, input_file_or_string, n_best=10, threshold=0.6, window=2):
        # load the content of the document.
        self.load_document(input=input_file_or_string, language='zh', normalization=None)
        self.clear_cache()
        # select {1-3}-grams not containing punctuation marks and not
        #    beginning/ending with a stop word as candidates.
        self.candidate_selection(n=3)
        # weight the candidates using YAKE weighting scheme, a window (in
        #    words) for computing left/right contexts can be specified.
        self.candidate_weighting(window=window)
        # Get the 10-highest scored candidates as keyphrases.
        #    redundant keyphrases are removed from the output using levenshtein
        #    distance and a threshold.
        keyphrases = self.get_n_best(n=n_best, threshold=threshold)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Seon
@description:

YAKE(Yet Another Keyword Extractor)，一种基于关键词统计的单文档无监督关键词提取算法：
基于5种指标：是否大写，词的位置，词频，上下文关系，词在句中频率，来计算候选词的得分，从而筛选Top-N关键词。
中文只用后4个指标。

modify from: https://pypi.org/project/iyake-cn/0.5.5/#files
"""
import os
import jieba
import pandas as pd
import re
from math import log10, sqrt
from jieba import lcut, posseg

pwd_path = os.path.abspath(os.path.dirname(__file__))

# inner data file
default_stopwords_path = os.path.join(pwd_path, '../data/stopwords.txt')
jieba.setLogLevel('ERROR')


def get_pos_lst(words_lst):
    # 分词的位置列表
    return list(zip(words_lst, range(1, len(words_lst) + 1)))


def get_T_pos(pos_lst, word, median_fn):
    # 单个词的T_pos指标
    _lst = [i[1] for i in pos_lst if word in i[0]]
    _lst.sort()
    half = len(_lst) // 2
    median = (_lst[half] + _lst[~half]) / 2
    if median_fn is None:
        return log10(log10(median + 10)) + 1
    if callable(median_fn):
        return median_fn(median)


def words_count(words_lst):
    # 全文词频统计，返回字典
    counts_dict = {}
    for word in words_lst:
        counts_dict[word] = counts_dict.get(word, 0) + 1
    return counts_dict


def related_content(words_lst, word, size):
    # 单个词的T_Rel指标
    split_lst = ' '.join(words_lst).split(word)
    left_lst = split_lst[:-1]
    right_lst = split_lst[1:]
    DL = 0
    for i in left_lst:
        # 从 word 出现的每个地方往左取 size 个词，计算不重复词数
        left_words = i.split(' ')[-2: -2 - size: -1]
        DL += len(set(left_words))

    DR = 0
    for i in right_lst:
        # 从 word 出现的每个地方往右
        right_words = i.split(' ')[1: size + 1]
        DR += len(set(right_words))

    return DL / len(left_lst) + DR / len(right_lst)


def get_pseg(x):
    # 词性
    return [p for w, p in list(posseg.cut(x))][0]


def std(lst):
    ex = float(sum(lst)) / len(lst)
    s = 0
    for i in lst:
        s += (i - ex) ** 2
    return sqrt(float(s) / len(lst))


def get_S_t(content, only_cn=False, stop=None, pos_type='s', median_fn=None, tf_normal='yake', adjust=1,
            r_size=10):
    # 各项指标原始得分表
    if only_cn:
        # 纯中文分词
        clean_str = re.sub(r'[^一-龟，；。？！（）“”]+', '', content)
    else:
        clean_str = re.sub(r'[0-9]+', '', content)

    jb_lst = [w for w in lcut(clean_str) if len(w) > 1]  # 分词过滤单字

    if stop is not None:
        # 停用词
        jb_lst = [w for w in jb_lst if w not in stop]

    uni_lst = list(set(jb_lst))
    uni_lst.sort(key=jb_lst.index)  # 固定顺序唯一词
    split_content = re.split(r'[，；。？！]', clean_str)  # 原文本按标点拆分句子列表

    # 位置 T_pos 得分表
    pos_lst = 0
    if pos_type == 'w':
        pos_lst = get_pos_lst(jb_lst)  # 计算全文词位置
    elif pos_type == 's':
        pos_lst = get_pos_lst(split_content)  # 计算含词的句子位置
    t_pos_scores = []
    for w in uni_lst:
        t_pos_scores.append((w, get_T_pos(pos_lst, w, median_fn)))

    # 全文词频 TF_norm 得分表
    wc_dic = words_count(jb_lst)
    mean_tf = sum(wc_dic.values()) / len(uni_lst)  # 词频均值
    std_tf = std(list(wc_dic.values()))  # 词频标准差
    max_tf = max(list(wc_dic.values()))
    min_tf = min(list(wc_dic.values()))
    if max_tf - min_tf == 0:
        tf_normal = 'yake'
    tf_norm_scores = []
    TF_norm = 0
    for w in uni_lst:
        if tf_normal == 'yake':
            TF_norm = wc_dic.get(w) / (mean_tf + std_tf)  # yake版归一化
        if tf_normal == 'mm':
            TF_norm = (wc_dic.get(w) - min_tf) / (max_tf - min_tf)  # max-min归一化
        tf_norm_scores.append((w, TF_norm))

    # 上下文 T_Rel 得分表
    t_rel_scores = []
    all_words = len(uni_lst)
    for w in uni_lst:
        DL_RL = related_content(jb_lst, w, r_size)
        T_Rel = 1 + DL_RL * wc_dic.get(w) / all_words
        t_rel_scores.append((w, T_Rel))

    # 句间词频 T_sentence 得分表
    len_content = len(split_content)
    SF_dic = {}
    for w in uni_lst:
        for sentence in split_content:
            if w in sentence:
                SF_dic[w] = SF_dic.get(w, 0) + 1
    t_sentence_scores = [(i[0], i[1] / len_content) for i in list(SF_dic.items())]

    # 重要性 S_t 总分表
    df_scores = pd.DataFrame({
        'word': uni_lst,
        'fre': [wc_dic.get(i) for i in uni_lst],
        't_pos': [i[1] for i in t_pos_scores],
        'tf_norm': [i[1] for i in tf_norm_scores],
        't_rel': [i[1] for i in t_rel_scores],
        't_sentence': [i[1] for i in t_sentence_scores]
    })
    df_scores['pseg'] = df_scores['word'].apply(get_pseg)
    df_scores.eval(f's_t = t_pos*t_rel / ({adjust} + (tf_norm + t_sentence)/t_rel)', inplace=True)
    return df_scores


def get_key_words(df_scores, n_best=10, sort_col='s_t', ascend=True, p=None):
    # 获取关键词列表，默认前10个，升序
    if p is None:
        p = {'n', 'a'}
    p = list(p)
    df_scores = df_scores[df_scores['pseg'].isin(p)]
    df_items = df_scores.sort_values(sort_col, ascending=ascend)
    print(df_items)
    word_scores = zip(df_items['word'].to_list(), df_items['s_t'].to_list())
    return list(word_scores)[:n_best]


def get_stopwords(txt_file):
    return set([line.strip() for line in open(txt_file, 'r', encoding='utf-8').readlines()])


class YakeZH:
    def __init__(self, stopwords_path=None):
        if stopwords_path is None:
            stopwords_path = default_stopwords_path
        self.stopwords = get_stopwords(stopwords_path)

    def extract(self, text, n_best=10):
        df = get_S_t(text, stop=self.stopwords)
        keyphrases = get_key_words(df, n_best=n_best)
        return keyphrases


if __name__ == '__main__':
    txt = '物流很快，服务也很好，还有售后回馈。外观很时尚并且超大视野'
    m = YakeZH()
    words = m.extract(txt)
    print(words)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
modify from: https://github.com/boudinfl/pke/blob/master/pke/readers.py
"""
import re

import jieba
import jieba.posseg as pseg
import pandas as pd
from loguru import logger
from pke_zh.data_structures import Document

jieba.setLogLevel('ERROR')


def load_document_frequency_file(input_file, delimiter='\t'):
    """Load a tsv (tab-separated-values) file containing document frequencies.
    Automatically detects if input file is compressed (gzip) by looking at its
    extension (.gz).

    Args:
        input_file (str): the input file containing document frequencies in
            csv format.
        delimiter (str): the delimiter used for separating term-document
            frequencies tuples, defaults to '\t'.

    Returns:
        dict: a dictionary of the form {term_1: freq}, freq being an integer.
    """

    df_freq = pd.read_csv(input_file, sep=delimiter)
    df_freq['frequence'] = df_freq['frequence'].map(int)
    frequencies = df_freq.set_index('word').to_dict()['frequence']

    # return the populated dictionary
    return frequencies


def flatten_tuples(tuple_list):
    """
    :param tuple_list: list of list, each list contains tuple elements.
    :return: flatten tuple list
    """
    result = [t for tlist in tuple_list for t in tlist]
    return result


def fetch_nested_list_elements(nested_list, idx):
    """
    :param nested_list: list of list, each list is a tuple
    :param idx: target index in a tuple
    :return: list of list, each list contains target elements
    """
    result = [[t[idx] for t in inner_list] for inner_list in nested_list]
    return result


def reduce_charoffset(list_str):
    len_list = [len(x) for x in list_str]
    cur_sum = 0
    res = [0]
    for i in range(len(len_list) - 1):
        cur_sum += len_list[i]
        res.append(cur_sum)
    offset = [(x, x + len_list[i]) for i, x in enumerate(res)]
    return offset


def cut_sent(para):
    para = re.sub('([。！？\?;；])([^”’])', r"\1\n\2", para)
    para = re.sub('(\.{6})([^”’])', r"\1\n\2", para)
    para = re.sub('(\…{2})([^”’])', r"\1\n\2", para)
    para = re.sub('([。！？；;\?][”’])([^，。！？\?])', r'\1\n\2', para)
    para = para.rstrip()
    paras = [p.strip() for p in para.split("\n") if p.strip()]
    return paras


class ChineseNLP(object):
    def __init__(self, max_length=1e6):
        self.max_length = max_length

    def __call__(self, text):
        """Split sentences
        Args:
            text(str): text for NLP Pipeline Processing.
        """
        clean_text = self.clean(text)
        sentences = cut_sent(clean_text)

        pseg_results, words, postags = self.sentence_pseg(sentences)

        char_offsets = list(map(lambda ws: reduce_charoffset(ws), words))

        # flatten char_offsets and make global index in the document.
        global_char_offsets = [char_offsets[0]]
        last_sent_offset = char_offsets[0][-1][1]
        for offsets in char_offsets[1:]:
            new_offset = [(x[0] + last_sent_offset, x[1] + last_sent_offset) for x in offsets]
            last_sent_offset = new_offset[-1][1]
            global_char_offsets.append(new_offset)
        global_char_offsets = flatten_tuples(global_char_offsets)

        result = [{'words': ws,
                   'POS': postags[i],
                   'char_offsets': char_offsets[i],
                   'global_char_offsets': global_char_offsets,
                   'word_POS': pseg_results}
                  for i, ws in enumerate(words)]
        return result

    def sentence_pseg(self, sentences):
        results = [[(s.word, s.flag) for s in pseg.cut(sent)] for sent in sentences]
        words = fetch_nested_list_elements(results, 0)
        postags = fetch_nested_list_elements(results, 1)
        return results, words, postags

    def clean(self, text):
        text = re.sub(r'img[_\d]+', '', text.strip())
        text = re.sub(r'_SEG_', '', text.strip())
        return text


class Reader(object):
    def read(self, path):
        raise NotImplementedError


class RawTextReader(Reader):
    """Reader for raw text."""

    def __init__(self, language='zh'):
        """Constructor for RawTextReader.

        Args:
            language (str): language of text to process.
        """

        self.language = language
        if self.language != 'zh':
            logger.warning('toolkit is created for zh language!')
        self.nlp = ChineseNLP()

    def read(self, text, **kwargs):
        """Read the input file and use spacy to pre-process.

        Args:
            text (str): raw text to pre-process.
            max_length (int): maximum number of characters in a single text for
                spacy, default to 1,000,000 characters (1mb).
        """
        sentences = self.nlp(text)
        doc = Document.from_sentences(sentences,
                                      input_file=kwargs.get('input_file', None),
                                      **kwargs)

        return doc



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

refer: https://maartengr.github.io/KeyBERT/guides/quickstart.html
keybert关键词抽取，核心思想类似embedrank，
    只是向量提取的度量方法使用keyBERT
"""

from loguru import logger
import itertools
from sklearn.metrics.pairwise import cosine_similarity
from text2vec import SentenceModel
import numpy as np
from typing import List

from pke_zh.base import BaseKeywordExtractModel


def max_sum_ranking(doc_embedding: np.ndarray,
                    can_embeddings: np.ndarray,
                    can_names: List[str],
                    top_n: int,
                    nr_candidates: int):
    """ Calculate Max Sum Distance for extraction of keywords
        We take the 2 x top_n most similar words/phrases to the document.
        Then, we take all top_n combinations from the 2 x top_n words and
        extract the combination that are the least similar to each other
        by cosine similarity.
        NOTE:
            This is O(n^2) and therefore not advised if you use a large top_n
        Arguments:
            doc_embedding: The document embeddings
            can_embeddings: The embeddings of the selected candidate keywords/phrases
            can_names: The selected candidate keywords/keyphrases
            top_n: 取top_n 个关键词
            nr_candidates: The number of candidates to consider, generaly set top_n *2
        Returns:
             List[Tuple[str, float]]: The selected keywords/keyphrases with their distances
        """
    # calculate distances and extract words
    # print(doc_embedding)
    distances = cosine_similarity(doc_embedding, can_embeddings)
    distance_words = cosine_similarity(can_embeddings)

    # Get 2*top_n words as candidates based on cosine similarity
    can_idx = list(distances.argsort()[0][-nr_candidates:])

    can_name_filter = [can_names[i] for i in can_idx]
    cand_distance = distance_words[np.ix_(can_idx, can_idx)]

    # Calculate the 候选词里的topn of words的组合， that are the least similar to each other
    min_sim = 100000
    final_candidate = None
    # print(can_idx)
    for combination in itertools.combinations(range(len(can_idx)), top_n):
        sim = sum([cand_distance[i][j] for i in combination for j in combination if i != j])
        if sim < min_sim:
            final_candidate = combination
            min_sim = sim
    # return candi_name and score
    result = []
    if not final_candidate:
        final_candidate = can_idx
    for val in final_candidate:
        result.append((can_name_filter[val], distances[0][can_idx[val]]))
    return result


def mmr_ranking(doc_embedding: np.ndarray,
                can_embeddings: np.ndarray,
                can_names: List[str],
                top_n: int,
                alpha: float = 0.5):
    """ Calculate Maximal Marginal Relevance (MMR)
    between candidate keywords and the document.
    MMR considers the similarity of keywords/keyphrases with the
    document, along with the similarity of already selected
    keywords and keyphrases. This results in a selection of keywords
    that maximize their within diversity with respect to the document.
    Arguments:
        doc_embedding: The document embeddings
        can_embeddings: The embeddings of the selected candidate keywords/phrases
        can_names: The selected candidate keywords/keyphrases
        top_n: The number of keywords/keyhprases to return
        alpha: How diverse the select keywords/keyphrases are.
                   Values between 0 and 1 with 0 being not diverse at all
                   and 1 being most diverse.
    Returns:
         List[Tuple[str, float]]: The selected keywords/keyphrases with their distances
    """
    # calculate distances and extract words
    doc_can_distances = cosine_similarity(can_embeddings, doc_embedding)
    distance_words = cosine_similarity(can_embeddings)

    # Initialize candidates and already choose best keyword/keyphras
    keywords_idx = [np.argmax(doc_can_distances)]
    candidates_idx = [i for i in range(len(can_names)) if i != keywords_idx[0]]

    for r in range(min(top_n, len(can_embeddings) - 1)):
        # extract similarities
        candidate_similarities = doc_can_distances[candidates_idx, :]
        target_similarities = np.max(distance_words[candidates_idx][:, keywords_idx], axis=1)

        # Calculate MMR
        mmr = alpha * candidate_similarities - (1 - alpha) * target_similarities.reshape(-1, 1)
        mmr_idx = candidates_idx[np.argmax(mmr)]

        # Update keywords & candidates
        keywords_idx.append(mmr_idx)
        candidates_idx.remove(mmr_idx)

    # return candia_name and score
    result = []
    for val in keywords_idx:
        result.append((can_names[val], doc_can_distances[val][0]))
    return result


def mmr_norm_ranking(doc_embedding: np.ndarray,
                     can_embeddings: np.ndarray,
                     can_names: List[str],
                     top_n: int,
                     alpha: float = 0.5):
    """Rank candidates according to a query

    :param document: np.array, dense representation of document (query)
    :param candidates: np.array, dense representation of candidates
    :param l: float, ratio between distance to query or distance between
        chosen candidates
    Returns: a list of candidates rank
    """

    def norm(sim, **kwargs):
        sim -= sim.min(**kwargs)
        sim /= (sim.max(**kwargs) + 1e-10)
        sim = 0.5 + (sim - sim.mean(**kwargs)) / (sim.std(**kwargs) + 1e-10)
        return sim

    def norm2(sim, **kwargs):
        min_ = sim.min(**kwargs)
        max_ = (sim.max(**kwargs) + 1e-10)
        sim = (sim - min_) / max_
        sim = 0.5 + (sim - sim.mean(**kwargs)) / (sim.std(**kwargs) + 1e-10)
        return sim

    sim_doc = cosine_similarity(doc_embedding, can_embeddings)
    sim_doc[np.isnan(sim_doc)] = 0.
    sim_doc = norm(sim_doc)
    sim_doc[np.isnan(sim_doc)] = 0.

    sim_can = cosine_similarity(can_embeddings)
    sim_can[np.isnan(sim_can)] = 0.
    sim_can = norm(sim_can, axis=1)
    sim_can[np.isnan(sim_can)] = 0.

    sel = np.zeros(len(can_embeddings), dtype=bool)
    ranks = [None] * len(can_embeddings)
    # Compute first candidate, the second part of the calculation is 0
    # as there are no other chosen candidates to maximise distance to
    chosen_candidate = (sim_doc * alpha).argmax()
    sel[chosen_candidate] = True
    ranks[chosen_candidate] = 0

    for r in range(1, len(can_embeddings)):
        # Remove already chosen candidates
        sim_can[sel] = np.nan

        # Compute MMR score
        scores = alpha * sim_doc - (1 - alpha) * sim_can[:, sel].max(axis=1)
        chosen_candidate = np.nanargmax(scores)

        # Update output and mask with chosen candidate
        sel[chosen_candidate] = True
        ranks[chosen_candidate] = r

    result = []
    for can_id, val in enumerate(ranks):
        if not val is None:
            result.append((can_names[can_id], (len(ranks) - 1 - val) / (len(ranks) - 1)))

    return result


class KeyBert(BaseKeywordExtractModel):
    def __init__(self, model='shibing624/text2vec-base-chinese'):
        """
            原文支持若干种embedding 方法：SentenceModel、SentenceTransformers、Flair、Spacy、gensim
            中文默认支持text2vec.SentenceModel model="shibing624/text2vec-base-chinese"模型，
            英文可设置model="sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2"模型，
            其他语言参考：sentence-transformers models:
                  * https://www.sbert.net/docs/pretrained_models.html
            """
        # param: model sentenceTransformers
        super(KeyBert, self).__init__()
        if isinstance(model, str):
            try:
                self.model = SentenceModel(model)
            except Exception as e:
                logger.error('wrong url for sentence model, change to default!')
                self.model = SentenceModel('shibing624/text2vec-base-chinese')
        elif isinstance(model, SentenceModel):
            self.model = model
        else:
            raise ValueError('model must be str or text2vec.SentenceModel')
        self.max_length = self.model.max_seq_length

    def candidate_selection(self, pos=None):
        """Candidate selection using longest sequences of PoS.
        :param pos: set of valid POS tags, defaults to ('NOUN', 'ADJ').
        """
        if pos is not None:
            self.valid_pos = pos

        # select sequence of adjectives and nouns
        self.longest_pos_sequence_selection(valid_pos=self.valid_pos)
        self.candidate_filtering()

    def _flatten_doc_words(self, lower):
        """flatten sentence words whose postags are valid"""
        doc = ' '.join(w.lower() if lower else w for s in self.sentences
                       for i, w in enumerate(s.words)
                       if s.pos[i] in self.valid_pos)
        return doc

    def _calculate_candidate_weights(self, rank, cand_name):
        for candidate_id, r in enumerate(rank):
            if len(rank) > 1:
                # Inverting ranks so the first ranked candidate has the biggest score
                score = (len(rank) - 1 - r) / (len(rank) - 1)
            else:
                score = r
            self.weights[cand_name[candidate_id]] = score

    def _doc_to_sent_list(self):
        """
        为了充分获取长文本语义，针对单句做的合并
        :return: sentences list
        """
        sentence = []
        cur_len = 0
        cur_sent = []
        for i, sent in enumerate(self.sentences):
            cur_text = ''.join(sent.words)
            cur_len += len(cur_text)
            if cur_len >= self.max_length and cur_sent:
                sentence.append(''.join(cur_sent))
                cur_sent = [cur_text]
                cur_len = len(cur_text)
            else:
                cur_sent.append(cur_text)
        if cur_len:
            sentence.append("".join(cur_sent))
        return sentence

    def _weights_update(self, canlist):
        for canname, score in canlist:
            self.weights[canname] = score

    def candidate_weighting(self, use_maxsum=True, use_mmr=False, top_n=10, alpha=0.5, nr_candidates=20):
        """Candidate weighting function using distance to document.

        :param l: float, Lambda parameter for EmbedRank++ Maximal Marginal
        Relevance (MMR) computation. Use 1 to compute EmbedRank and 0 to not
        use the document, but only the most diverse set of candidates
        (defaults to 1).
        """
        # get doc's sentences
        doc_sents = self._doc_to_sent_list()
        doc_embed = self.model.encode(doc_sents)
        doc_embed = np.average(doc_embed, axis=0)  # 取平均
        doc_embed = np.expand_dims(doc_embed, axis=0)  # 增加一个维度

        cand_name = list(self.candidates.keys())  # 记得是带空格的
        cand_embed = self.model.encode(cand_name)

        if use_mmr:
            can_list = mmr_ranking(doc_embed, cand_embed, cand_name, top_n, alpha)
            self._weights_update(can_list)
        elif use_maxsum:
            can_list = max_sum_ranking(doc_embed, cand_embed, cand_name, top_n, nr_candidates)
            self._weights_update(can_list)
        else:
            distances = cosine_similarity(doc_embed, cand_embed)
            for i, score in enumerate(distances[0]):
                self.weights[cand_name[i]] = score

    def extract(self, input_file_or_string, n_best=10, use_maxsum=True, use_mmr=False, alpha=0.5, nr_candidates=20):
        """
        提取text的关键词
        :param input_file_or_string: 输入doc
        :param n_best: top n_best个关键词
        :param use_maxsum: 是否使用maxsum similarity for the selection of keywords
        :param use_mmr: Whether to use Maximal Marginal Relevance (MMR) for the selection of keywords/keyphrases
        :param alpha: mmr算法的超参数，if use_mmr is set True, default:0.5
        :param nr_candidates: The number of candidates to consider if use_maxsum is set to True
        :return: keywords list
        """
        # 1. load the content of the document.
        self.load_document(input=input_file_or_string, language='zh', normalization=None)
        # 2. select sequences of nouns and adjectives as candidates.
        self.candidate_selection()
        # 3. weight the candidates using EmbedRank method
        self.candidate_weighting(use_maxsum, use_mmr, n_best, alpha, nr_candidates)
        # 4. get the 10-highest scored candidates as keyphrases
        keyphrases = self.get_n_best(n=n_best, redundancy_removal=True)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

TextRank keyphrase extraction model.

Implementation of the TextRank model for keyword extraction described in:

* Rada Mihalcea and Paul Tarau.
  TextRank: Bringing Order into Texts
  *In Proceedings of EMNLP*, 2004.

"""

import math
import networkx as nx

from pke_zh.base import BaseKeywordExtractModel


class TextRank(BaseKeywordExtractModel):
    """TextRank for keyword extraction.

    This model builds a graph that represents the text. A graph based ranking
    algorithm is then applied to extract the lexical units (here the words) that
    are most important in the text.

    In this implementation, nodes are words of certain part-of-speech (nouns
    and adjectives) and edges represent co-occurrence relation, controlled by
    the distance between word occurrences (here a window of 2 words). Nodes
    are ranked by the TextRank graph-based ranking algorithm in its unweighted
    variant.
    """

    def __init__(self):
        """Redefining initializer for TextRank."""
        super(TextRank, self).__init__()

        self.graph = nx.Graph()
        """The word graph."""

    def clear_cache(self):
        self.graph.clear()

    def candidate_selection(self, pos=None):
        """Candidate selection using longest sequences of PoS.
        :param pos: set of valid POS tags, defaults to ('NOUN', 'ADJ').
        """
        if pos is None:
            pos = {'n', 'a'}

        # select sequence of adjectives and nouns
        self.longest_pos_sequence_selection(valid_pos=pos)

    def calculate_bi_nodes_weights(self, wi, window, text, node1):
        for wj in range(wi + 1, min(wi + window, len(text))):
            node2, is_in_graph2 = text[wj]
            if is_in_graph2 and node1 != node2:
                self.graph.add_edge(node1, node2)

    def _flatten_words_and_add_nodes(self, pos):
        text = [(word, sentence.pos[i] in pos) for sentence in self.sentences
                for i, word in enumerate(sentence.words)]

        # add nodes to the graph
        self.graph.add_nodes_from([word for word, valid in text if valid])

        return text

    def build_word_graph(self, window=2, pos=None):
        """Build a word co-occurrence graph representation of the document

        Syntactic filters can be applied to select only words of certain Part-of-Speech.
        Co-occurrence relations can be controlled using the distance between
        word occurrences in the document.

        As the original paper does not give precise details on how the word
        graph is constructed, we make the following assumptions from the example
        given in Figure 2: 1) sentence boundaries **are not** taken into account
        and, 2) stopwords and punctuation marks **are** considered as words when
        computing the window.

        :param window: int, the window for connecting two words in the graph,
            defaults to 2.
        :param pos: the set of valid pos for words to be considered as nodes
            in the graph, defaults to ('NOUN'', 'ADJ').
        """
        if pos is None:
            pos = {'n', 'a'}
        # flatten document as a sequence of (word, pass_syntactic_filter) tuples
        text = self._flatten_words_and_add_nodes(pos)

        # add edges to the graph
        for i, (node1, is_in_graph1) in enumerate(text):
            # speed up things
            if not is_in_graph1:
                continue
            self.calculate_bi_nodes_weights(i, window, text, node1)

    def candidate_weighting(self,
                            window=2,
                            pos=None,
                            top_percent=None,
                            normalized=False):
        """Tailored candidate ranking method for TextRank.

        Keyphrase candidates are either composed from the T-percent highest-ranked words as in the
        original paper or extracted using the `candidate_selection()` method.
        Candidates are ranked using the sum of their (normalized?) words.

        :param window: int, the window for connecting two words in the graph,
            defaults to 2.
        :param pos: the set of valid pos for words to be considered as nodes
            in the graph, defaults to ('NOUN', 'PROPN', 'ADJ').
        :param top_percent: float, percentage of top vertices to keep for phrase
            generation.
        :param normalized: bool, normalize keyphrase score by their length,
            defaults to False.
        """
        if pos is None:
            pos = {'n', 'a'}

        # build the word graph
        self.build_word_graph(window=window, pos=pos)

        # compute the word scores using the unweighted PageRank formulae
        word_scores = nx.pagerank_scipy(self.graph, alpha=0.85, tol=0.0001, weight=None)

        # generate the phrases from the T-percent top ranked words
        if top_percent is not None:
            # computing the number of top keywords
            nb_nodes = self.graph.number_of_nodes()
            to_keep = min(math.floor(nb_nodes * top_percent), nb_nodes)
            to_keep = max(1, to_keep)

            # sorting the nodes by decreasing scores
            top_words = sorted(word_scores, key=word_scores.get, reverse=True)

            # creating keyphrases from the T-top words
            self.longest_keyword_sequence_selection(top_words[:int(to_keep)])
        # weight candidates using the sum of their word scores
        self.candidate_weight_calculate(word_scores, normalized)

    def candidate_weight_calculate(self, word_scores, normalized):
        for k in self.candidates.keys():
            tokens = self.candidates[k].lexical_form
            self.weights[k] = sum([word_scores[t] for t in tokens])
            if normalized:
                self.weights[k] /= len(tokens)
            # use position to break ties
            self.weights[k] += (self.candidates[k].offsets[0] * 1e-8)

    def extract(self, input_file_or_string, n_best=10, pos=None, top_percent=0.33):
        self.load_document(input=input_file_or_string,
                           language='zh',
                           normalization=None)

        # clear previous graph is exists
        self.clear_cache()

        self.candidate_weighting(
            window=2,
            pos=pos,
            top_percent=top_percent
        )
        keyphrases = self.get_n_best(n=n_best)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import os
from pke_zh.version import __version__

USER_DATA_DIR = os.path.expanduser('~/.cache/pke_zh/')
os.makedirs(USER_DATA_DIR, exist_ok=True)

from pke_zh import unsupervised
from pke_zh import supervised



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

TF-IDF keyphrase extraction model.

"""

from jieba.analyse.tfidf import DEFAULT_IDF
from pke_zh.base import BaseKeywordExtractModel


class TfIdf(BaseKeywordExtractModel):
    """TF*IDF keyphrase extraction model."""

    def __init__(self, idf_path=None, stopwords_path=None):
        """Redefining initializer for TopicRank."""
        super(TfIdf, self).__init__(stopwords_path)
        self.path = ""
        self.idf_freq = {}
        self.median_idf = 0.0
        idf_path = idf_path if idf_path is not None else DEFAULT_IDF
        self.set_new_path(idf_path)

    def set_new_path(self, new_idf_path):
        if self.path != new_idf_path:
            self.path = new_idf_path
            content = open(new_idf_path, 'r', encoding='utf-8').read()
            self.idf_freq = {}
            for line in content.splitlines():
                word, freq = line.strip().split(' ')
                self.idf_freq[word] = float(freq)
            self.median_idf = sorted(
                self.idf_freq.values())[len(self.idf_freq) // 2]

    def get_idf(self):
        return self.idf_freq, self.median_idf

    def candidate_selection(self, n=3, stoplist=None):
        """Select 1-3 grams as keyphrase candidates.

        :param n: int, the length of the n-grams, defaults to 3.
        :param stoplist: the stoplist for filtering candidates, defaults to
            `None`. Words that are punctuation marks from
            `string.punctuation` are not allowed.
        """
        # select ngrams from 1 to 3 grams
        self.ngram_selection(n=n)

        # initialize empty list if stoplist is not provided
        if stoplist is not None:
            self.stoplist = stoplist

        # filter candidates containing punctuation marks
        self.candidate_filtering()

    def candidate_weighting(self):
        """Candidate weighting function using document frequencies."""
        total = sum([len(v.surface_forms) for v in self.candidates.values()])

        for k, v in self.candidates.items():
            # add the idf score to the weights container
            self.weights[k] = len(v.surface_forms) / total * self.idf_freq.get(k, self.median_idf)

    def extract(self, input_file_or_string, n_best=10, pos=None):
        self.load_document(input=input_file_or_string, language='zh')
        self.candidate_selection(n=3)
        self.candidate_weighting()
        keyphrases = self.get_n_best(n=n_best, redundancy_removal=True)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

import re
import os
from string import punctuation
from loguru import logger
import codecs
from collections import defaultdict
from six import string_types

from pke_zh.data_structures import Candidate, Document
from pke_zh.readers import RawTextReader

pwd_path = os.path.abspath(os.path.dirname(__file__))
# inner data file
default_stopwords_path = os.path.join(pwd_path, 'data/stopwords.txt')

map_tag = {'ad': 'a', 'ag': 'a', 'an': 'a',
           'ng': 'n', 'nr': 'n', 'nrfg': 'n', 'nrt': 'n', 'ns': 'n', 'nt': 'n', 'nz': 'n',
           'rg': 'r', 'rr': 'r', 'rz': 'r',
           }


def load_stopwords(file_path):
    stopwords = set()
    if file_path and os.path.exists(file_path):
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                stopwords.add(line)
    return stopwords


class BaseKeywordExtractModel(object):
    """The ket class that provides base functions."""

    def __init__(self, stopwords_path='', valid_pos=None, self_defined_keyword_path=None):
        """Initializer for base class."""

        self.input_file = None
        """Path to the input file."""

        self.language = None
        """Language of the input file."""

        self.normalization = None
        """Word normalization method."""

        self.sentences = []
        """Sentence container (list of Sentence objects)."""

        self.candidates = defaultdict(Candidate)
        """Keyphrase candidates container (dict of Candidate objects)."""

        self.weights = {}
        """Weight container (can be either word or candidate weights)."""

        self.stoplist = list(load_stopwords(stopwords_path)) if stopwords_path and os.path.exists(stopwords_path) \
            else list(load_stopwords(default_stopwords_path))
        """List of stopwords."""

        self.valid_pos = {'n', 'a'} if valid_pos is None else valid_pos

        punctuation_expand = ['/', ',', '$', '%', '^', '*', '(', '+', '"', "'", ']', '+', '|', '[', '+', '——', '！',
                              '，', '、', '~', '@', '#', '￥', '%', '&', '*', '（', '）', '：', '；', '《', '）', '《',
                              '》', '“', '”', '(', ')', '»', '〔', '〕', '-']
        self.punctuations = list(punctuation) + punctuation_expand

        self.self_defined_keywords = []
        if self_defined_keyword_path is not None:
            words = open(self_defined_keyword_path, 'r', encoding='utf-8').readlines()
            words = [x.strip for x in words if x.strip()]
            words = sorted(words, key=lambda w: len(w), reverse=True)
            if len(words) > 10000:
                logger.warning(
                    "Too much matching-keyword! We recommend you try the HugeKeywordMatching algo.")
            self.self_defined_keywords.extend(words)

        self.raw_text = ""

    def clear_cache(self):
        raise NotImplementedError

    def load_document(self, input, **kwargs):
        """Loads the content of a document/string/stream in a given language.

        :param input: str, input.
        :param language: str, language of the input, defaults to 'en'.
        :param encoding: str, encoding of the raw file.
        :param normalization: str, word normalization method, defaults to
            'stemming'. Other possible values are 'lemmatization' or 'None'
            for using word surface forms instead of stems/lemmas.
        """
        # get the language parameter
        language = kwargs.get('language', 'zh')

        # test whether the language is known, otherwise fall back to english
        kwargs['language'] = language

        # initialize document
        doc = Document()
        self.raw_text = ""

        if isinstance(input, string_types):
            # if input is an input file
            if os.path.isfile(input):
                # other extensions are considered as raw text
                parser = RawTextReader(language=language)
                encoding = kwargs.get('encoding', 'utf-8')
                with codecs.open(input, 'r', encoding=encoding) as file:
                    self.raw_text = file.read()
                doc = parser.read(text=self.raw_text, path=input, **kwargs)

            # if input is a string
            else:
                parser = RawTextReader(language=language)
                doc = parser.read(text=input, **kwargs)
                self.raw_text = input
        # set the input file
        self.input_file = doc.input_file

        # set the language of the document
        self.language = language

        # set the sentences
        self.sentences = doc.sentences

        # POS normalization
        self.normalize_pos_tags()

        # clear previous candidates and weights if exists
        self.candidates.clear()
        self.weights.clear()

    def normalize_pos_tags(self):
        """Normalizes some jieba's postags."""
        for i, sentence in enumerate(self.sentences):
            self.sentences[i].pos = [map_tag.get(tag, tag) for tag in sentence.pos]

    @staticmethod
    def is_redundant(candidate, prev, minimum_length=1):
        """Test if one candidate is redundant

        A candidate is considered redundant if it is
        included in another candidate that is ranked higher in the list.

        :param candidate: str, the lexical form of the candidate.
        :param prev: list, the list of already selected candidates (lexical forms).
        :param minimum_length: int, minimum length (in words) of the candidate
            to be considered, defaults to 1.
        """
        # get the tokenized lexical form from the candidate
        candidate = re.sub(r'\s', '', candidate)

        # only consider candidate greater than one word
        if len(candidate) < minimum_length:
            return False

        # get the tokenized lexical forms from the selected candidates
        prev = [re.sub(r'\s', '', u) for u in prev]

        # loop through the already selected candidates
        redundant_status = any([prev_candidate in candidate or candidate in prev_candidate
                                for prev_candidate in prev])

        return redundant_status

    def redundancy_removal_best(self, best, num_candidates):
        # initialize a new container for non redundant candidates
        non_redundant_best = []

        # loop through the best candidates
        for candidate in best:
            # test weather candidate is redundant
            if self.is_redundant(candidate, non_redundant_best):
                continue

            # add the candidate otherwise
            non_redundant_best.append(candidate)

            # break computation if the n-best are found
            if len(non_redundant_best) >= num_candidates:
                break

        # copy non redundant candidates in best container
        return non_redundant_best

    def n_best_surface_form(self, stemming, best, top_n, n_best):
        if not stemming and isinstance(self.candidates, dict):
            n_best = [(' '.join(self.candidates[u].surface_forms[0]).lower(),
                       self.weights[u]) for u in best[:min(top_n, len(best))]]
        return n_best

    def get_n_best(self, n=10, redundancy_removal=True, stemming=False):
        """Returns the n-best candidates given the weights.

        :param n: int, the number of candidates, defaults to 10.
        :param redundancy_removal: bool, whether redundant keyphrases are
            filtered out from the n-best list, defaults to False.
        :param stemming: bool, whether to extract stems or surface forms
            (lowercased, first occurring form of candidate), default to
            False.
        """
        # sort candidates by descending weight
        best = sorted(self.weights, key=self.weights.get, reverse=True)

        # remove redundant candidates
        if redundancy_removal:
            best = self.redundancy_removal_best(best, n)

        # get the list of best candidates as (lexical form, weight) tuples
        n_best = [(u, self.weights[u]) for u in best[:min(n, len(best))]]

        # replace with surface forms if no stemming
        # when used in KPRank, self.candidates is a list.
        n_best = self.n_best_surface_form(stemming, best, n, n_best)

        # remove blank space of keyphrases
        n_best = [(re.sub(r'\s', '', u), weight) for u, weight in n_best]

        # if len(n_best) < n:
        #     logger.debug(
        #         'Not enough candidates to choose from '
        #         '({} requested, {} given)'.format(n, len(n_best)))

        # return the list of best candidates
        return n_best

    def add_candidate(self, words, pos, offset, sentence_id):
        """Add a keyphrase candidate to the candidates container.

        :param words: list, the words (surface form) of the candidate.
        :param pos: list, the Part-Of-Speeches of the words in the candidate.
        :param offset: int, the offset of the first word of the candidate.
        :param sentence_id: int, the sentence id of the candidate.
        """
        # build the lexical (canonical) form of the candidate using stems
        lexical_form = ' '.join(words)

        # add/update the surface forms
        self.candidates[lexical_form].surface_forms.append(words)

        # add/update the lexical_form
        self.candidates[lexical_form].lexical_form = words

        # add/update the POS patterns
        self.candidates[lexical_form].pos_patterns.append(pos)

        # add/update the offsets
        self.candidates[lexical_form].offsets.append(offset)

        # add/update the sentence ids
        self.candidates[lexical_form].sentence_ids.append(sentence_id)

    def ngram_selection(self, n=3):
        """Select all the n-grams and populate the candidate container.

        :param n: int, the n-gram length, defaults to 3.
        """
        # loop through the sentences
        for i, sentence in enumerate(self.sentences):
            # limit the maximum n for short sentence
            skip = min(n, sentence.length)
            # compute the offset shift for the sentence
            shift = sum([s.length for s in self.sentences[0:i]])
            # generate the ngrams
            for j in range(sentence.length):
                for k in range(j + 1, min(j + 1 + skip, sentence.length + 1)):
                    # add the ngram to the candidate container
                    self.add_candidate(words=sentence.words[j:k],
                                       # stems=sentence.stems[j:k],
                                       pos=sentence.pos[j:k],
                                       offset=shift + j,
                                       sentence_id=i)

    def longest_pos_sequence_selection(self, valid_pos=None):
        self.longest_sequence_selection(
            key=lambda s: s.pos, valid_values=valid_pos)

    def longest_keyword_sequence_selection(self, keywords):
        self.longest_sequence_selection(
            key=lambda s: s.words, valid_values=keywords)

    def longest_sequence_selection(self, key, valid_values):
        """Select the longest sequences of given POS tags as candidates.

        :param key: function that given a sentence return an iterable
        :param valid_values: set, the set of valid values, defaults to None.
        """
        # loop through the sentences
        for i, sentence in enumerate(self.sentences):
            # compute the offset shift for the sentence
            shift = sum(list(map(lambda s: s.length, self.sentences[0:i])))
            # container for the sequence (defined as list of offsets)
            seq = []

            # loop through the tokens
            for j, value in enumerate(key(self.sentences[i])):
                # add candidate offset in sequence and continue if not last word
                if value in valid_values:
                    seq.append(j)
                    if j < (sentence.length - 1):
                        continue

                # add sequence as candidate if non empty
                self.add_seq_as_candidates(seq, sentence, shift, i, j)

                # flush sequence container
                seq = []

    def add_seq_as_candidates(self, seq, sentence, shift, i, j):
        if len(seq) == 0:
            return

        # bias for candidate in last position within sentence
        bias = 0
        if j == (sentence.length - 1):
            bias = 1

        # add the ngram to the candidate container
        self.add_candidate(words=sentence.words[seq[0]:seq[-1] + 1],
                           pos=sentence.pos[seq[0]:seq[-1] + 1],
                           offset=shift + j - len(seq) + bias,
                           sentence_id=i)

    @staticmethod
    def _is_alphanum(word, valid_punctuation_marks='-'):
        """Check if a word is valid, i.e. it contains only alpha-numeric
        characters and valid punctuation marks.

        :param word: str, a word.
        :param valid_punctuation_marks: str, punctuation marks that are valid
                for a candidate, defaults to '-'.
        """
        for punct in valid_punctuation_marks.split():
            word = word.replace(punct, '')
        return word.isalnum()

    @staticmethod
    def discard_length_invalid_candidate(k, v, words, minimum_length, minimum_word_size, maximum_word_number):
        # discard candidates composed of 1-2 characters
        if len(''.join(words)) < minimum_length:
            return k

        # discard candidates containing small words (1-character)
        elif min([len(u) for u in words]) < minimum_word_size:
            return k

        # discard candidates composed of more than 5 words
        elif len(v.lexical_form) > maximum_word_number:
            return k
        return ''

    def discard_candidate(self, k, v, stoplist=None, minimum_length=2, minimum_word_size=2, valid_punctuation_marks='-',
                          maximum_word_number=7, pos_blacklist=None):
        # get the words from the first occurring surface form
        words = [u.lower() for u in v.surface_forms[0]]

        # discard if words are in the stoplist
        if stoplist is None:
            stoplist = set()
        if set(words).intersection(stoplist):
            return k
        # discard if tags are in the pos_blacklist
        elif set(v.pos_patterns[0]).intersection(pos_blacklist):
            return k
        # discard if containing tokens composed of only punctuation
        elif any([set(u).issubset(set(self.punctuations)) for u in words]):
            return k
        r = self.discard_length_invalid_candidate(k, v, words, minimum_length, minimum_word_size,
                                                  maximum_word_number)
        return r

    def candidate_filtering(self,
                            stoplist=None,
                            minimum_length=2,
                            minimum_word_size=2,
                            valid_punctuation_marks='-',
                            maximum_word_number=7,
                            pos_blacklist=None):
        """Filter the candidates containing strings from the stoplist.

        Only keep the candidates containing alpha-numeric characters (if the
        non_latin_filter is set to True) and those length exceeds a given
        number of characters.

        :param stoplist: list of strings, defaults to None.
        :param minimum_length: int, minimum number of characters for a
            candidate, defaults to 3.
        :param minimum_word_size: int, minimum number of characters for a
            token to be considered as a valid word, defaults to 2.
        :param valid_punctuation_marks: str, punctuation marks that are valid
            for a candidate, defaults to '-'.
        :param maximum_word_number: int, maximum length in words of the
            candidate, defaults to 5.
        :param pos_blacklist: list of unwanted Part-Of-Speeches in
            candidates, defaults to [].
        """
        if pos_blacklist is None:
            pos_blacklist = []

        del_keys = set([self.discard_candidate(k, self.candidates[k], stoplist, minimum_length, minimum_word_size,
                                               valid_punctuation_marks, maximum_word_number, pos_blacklist) for k in
                        list(self.candidates)])
        for k in list(self.candidates):
            if k in del_keys:
                del self.candidates[k]

    def self_defined_keyword_matching(self):
        """

        :param docs: list of doc texts to be extracted
        :param keywords: self-defined keyword set
        :return:
        """
        result = []
        for w in self.self_defined_keywords:
            if w in self.raw_text:
                result.append(w)
        return result

    def extract(self, input_file_or_string, n_best=10, **kwargs):
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 配置切词器
"""
import re
import jieba
from jieba import posseg

jieba.setLogLevel(log_level="ERROR")


def word_segment(sentence, cut_type='word', pos=False):
    """
    Word segmentation
    :param sentence:
    :param cut_type: 'word' use jieba.lcut; 'char' use list(sentence)
    :param pos: enable POS
    :return: list
    """
    if pos:
        if cut_type == 'word':
            word_pos_seq = posseg.lcut(sentence)
            word_seq, pos_seq = [], []
            for w, p in word_pos_seq:
                word_seq.append(w)
                pos_seq.append(p)
            return word_seq, pos_seq
        elif cut_type == 'char':
            word_seq = list(sentence)
            pos_seq = []
            for w in word_seq:
                w_p = posseg.lcut(w)
                pos_seq.append(w_p[0].flag)
            return word_seq, pos_seq
    else:
        if cut_type == 'word':
            return jieba.lcut(sentence)
        elif cut_type == 'char':
            return list(sentence)


def sentence_segment(text, delimiters=('？', '！', '。', '；', '……', '…'), include_symbols=False):
    """
    Sentence segmentation
    :param text: query
    :param delimiters: set
    :param include_symbols: bool
    :return: list(word, idx)
    """
    result = []
    delimiters = set([item for item in delimiters])
    delimiters_str = '|'.join(delimiters)
    blocks = re.split(delimiters_str, text)
    start_idx = 0
    for blk in blocks:
        if not blk:
            continue
        result.append((blk, start_idx))
        start_idx += len(blk)
        if include_symbols and start_idx < len(text):
            result.append((text[start_idx], start_idx))
            start_idx += 1
    return result


if __name__ == '__main__':
    text = "这个消息在北京城里不胫而走。你听过了吗？"
    words = '我，来。上海？吃？上海菜'
    wordlist = re.split('？|。', words)
    print(wordlist)

    print(text)
    t = word_segment(text, pos=True)
    print(t)
    t = sentence_segment(text, include_symbols=True)
    print(t)
    t = sentence_segment('这个消息在北京城。', include_symbols=True)
    print(t)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
import json
import os
import pickle


def load_pkl(pkl_path):
    """
    加载词典文件
    :param pkl_path:
    :return:
    """
    with open(pkl_path, 'rb') as f:
        result = pickle.load(f)
    return result


def save_pkl(vocab, pkl_path, overwrite=True):
    """
    存储文件
    :param pkl_path:
    :param overwrite:
    :return:
    """
    if os.path.exists(pkl_path) and not overwrite:
        return
    with open(pkl_path, 'wb') as f:
        pickle.dump(vocab, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_json(json_path, encoding='utf-8'):
    with open(json_path, mode='r', encoding=encoding) as json_file:
        data = json.load(json_file)
    return data


def save_json(data, json_path, mode='w', encoding='utf-8'):
    dir = os.path.dirname(os.path.abspath(json_path))
    if not os.path.exists(dir):
        print(dir)
        os.makedirs(dir)
    with open(json_path, mode=mode, encoding=encoding) as f:
        f.write(json.dumps(data, ensure_ascii=False))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 汉字处理的工具，判断unicode是否是汉字，数字，英文，或者其他字符
"""

import six


def convert_to_unicode(text):
    """Converts `text` to Unicode (if it's not already), assuming utf-8 input."""
    if six.PY3:
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode("utf-8", "ignore")
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    elif six.PY2:
        if isinstance(text, str):
            return text.decode("utf-8", "ignore")
        elif isinstance(text, unicode):
            return text
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))
    else:
        raise ValueError("Not running on Python2 or Python 3?")


def is_chinese(uchar):
    """判断一个unicode是否是汉字"""
    return '\u4e00' <= uchar <= '\u9fa5'


def is_chinese_string(string):
    """判断是否全为汉字"""
    return all(is_chinese(c) for c in string)


def is_number(uchar):
    """判断一个unicode是否是数字"""
    return u'u0030' <= uchar <= u'u0039'


def is_number_string(string):
    """判断是否全部是数字"""
    return all(is_number(c) for c in string)


def is_alphabet(uchar):
    """判断一个unicode是否是英文字母"""
    return u'u0041' <= uchar <= u'u005a' or u'u0061' <= uchar <= u'u007a'


def is_alphabet_string(string):
    """判断是否全部为英文字母"""
    return all('a' <= c <= 'z' for c in string)


def is_other(uchar):
    """判断是否非汉字，数字和英文字符"""
    return not (is_chinese(uchar) or is_number(uchar) or is_alphabet(uchar))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""

import sys

from setuptools import setup, find_packages

__version__ = None
exec(open('pke_zh/version.py').read())

if sys.version_info < (3,):
    sys.exit('Sorry, Python3 is required for pke_zh.')

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='pke_zh',
    version=__version__,
    description='pke_zh, context-aware bag-of-words term weights for query and document.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='XuMing',
    author_email='xuming624@qq.com',
    url='https://github.com/shibing624/pke_zh',
    license="Apache 2.0",
    zip_safe=False,
    python_requires='>=3.5',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: Chinese (Traditional)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    keywords='pke_zh,term weighting,textrank,word rank,wordweight',
    install_requires=[
        "loguru",
        "jieba",
        "pandas",
        "numpy",
        "six",
        "tqdm",
        "requests",
        "scikit-learn",
        "scipy",
        "networkx",
        "text2vec",
    ],
    packages=find_packages(exclude=['tests']),
    package_dir={'pke_zh': 'pke_zh'},
    package_data={'pke_zh': ['*.*', 'data/*', ]}
)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

SingleRank keyphrase extraction model.

Simple extension of the TextRank model described in:

* Xiaojun Wan and Jianguo Xiao.
  CollabRank: Towards a Collaborative Approach to Single-Document Keyphrase
  Extraction.
  *In proceedings of the COLING*, pages 969-976, 2008.
"""
import networkx as nx
from pke_zh.unsupervised.textrank import TextRank


class SingleRank(TextRank):
    """SingleRank keyphrase extraction model.

    This model is an extension of the TextRank model that uses the number of
    co-occurrences to weigh edges in the graph.
    """

    def __init__(self):
        """Redefining initializer for SingleRank."""
        super(SingleRank, self).__init__()

    def calculate_bi_nodes_weights(self, word_idx, window, text, node1):
        for j in range(word_idx + 1, min(word_idx + window, len(text))):
            node2, is_in_graph2 = text[j]
            if is_in_graph2 and node1 != node2:
                if not self.graph.has_edge(node1, node2):
                    self.graph.add_edge(node1, node2, weight=0.0)
                self.graph[node1][node2]['weight'] += 1.0

    def extract_word_and_pos_from_sentences(self, valid_pos):
        text = [(word, sentence.pos[i] in valid_pos) for sentence in self.sentences
                for i, word in enumerate(sentence.words)]
        return text

    def build_word_graph(self, window=10, pos=None):
        """Build a co-occurrence word graph representation of the document

        Syntactic filters can be applied to select only words of certain Part-of-Speech.
        Co-occurrence relations can be controlled using the distance (window) between word
        occurrences in the document. The number of times two words co-occur in a window is
        encoded as *edge weights*. Sentence boundaries **are not** taken into account in the window.

        :param window: int, the window for connecting two words in the graph,
            defaults to 10.
        :param pos: the set of valid pos for words to be considered as nodes
            in the graph, defaults to ('NOUN', 'ADJ').
        """
        if pos is None:
            pos = {'n', 'a'}

        # flatten document as a sequence of (word, pass_syntactic_filter) tuples
        text = self.extract_word_and_pos_from_sentences(pos)

        self.clear_cache()

        # add nodes to the graph
        valid_text = list(filter(lambda x: x[1], text))
        self.graph.add_nodes_from([word for word, _ in valid_text])
        # add edges to the graph
        for i, (node1, is_in_graph1) in enumerate(text):
            # speed up things
            if not is_in_graph1:
                continue

            self.calculate_bi_nodes_weights(i, window, text, node1)

    def candidate_weighting(self, window=10, pos=None, top_percent=None, normalized=False):
        """Keyphrase candidate ranking using the weighted variant of the TextRank formulae.

        Candidates are scored by the sum of the scores of their words.

        :param window: int, the window within the sentence for connecting two
            words in the graph, defaults to 10.
        :param pos: the set of valid pos for words to be considered as nodes
            in the graph, defaults to ('NOUN', 'PROPN', 'ADJ').
        :param normalized: bool, normalize keyphrase score by their length,
            defaults to False.
        """
        if pos is None:
            pos = {'n', 'a'}

        # build the word graph
        self.build_word_graph(window=window, pos=pos)

        # compute the word scores using random walk
        w = nx.pagerank_scipy(self.graph,
                              alpha=0.85,
                              tol=0.0001,
                              weight='weight')

        # loop through the candidates
        self.candidate_weight_calculate(w, normalized)

    def extract(self, input_file_or_string, n_best=10, pos=None):
        self.load_document(input=input_file_or_string, language='zh', normalization=None)
        self.candidate_selection(pos=pos)
        self.candidate_weighting(window=10, pos=pos)
        keyphrases = self.get_n_best(n=n_best)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com), Florian Boudin
@description:

Multipartite graph keyphrase extraction model.

Graph-based ranking approach to keyphrase extraction described in:

* Florian Boudin.
  Unsupervised Keyphrase Extraction with Multipartite Graphs.
  *In proceedings of NAACL*, pages 667-672, 2018.

"""
import math
from itertools import combinations

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist

from pke_zh.unsupervised.topicrank import TopicRank


class MultipartiteRank(TopicRank):
    """Multipartite graph keyphrase extraction model."""

    def __init__(self, stopwords_path=None):
        """Redefining initializer for MultipartiteRank."""
        super(MultipartiteRank, self).__init__(stopwords_path)

        self.topic_identifiers = {}
        """ A container for linking candidates to topic identifiers. """

        self.graph = nx.DiGraph()
        """ Redefine the graph as a directed graph. """

    def clear_cache(self):
        self.graph.clear()
        self.topics.clear()
        self.topic_identifiers.clear()

    def topic_clustering(self, threshold=0.74, method='average'):
        """Clustering candidates into topics.

        :param threshold: float, the minimum similarity for clustering,
            defaults to 0.74, i.e. more than 1/4 of stem overlap
            similarity.
        :param method: str, the linkage method, defaults to average.
        """
        # handle document with only one candidate
        if len(self.candidates) == 1:
            candidate = list(self.candidates)[0]
            self.topics.append([candidate])
            self.topic_identifiers[candidate] = 0
            return

        # vectorize the candidates
        candidates, distance_matrix = self.vectorize_candidates()

        # compute the distance matrix
        jaccard_matrix = pdist(distance_matrix, 'jaccard')
        jaccard_matrix = np.nan_to_num(jaccard_matrix)

        # compute the clusters
        cluster_matrix = linkage(jaccard_matrix, method=method)

        # form flat clusters
        clusters = fcluster(cluster_matrix, t=threshold, criterion='distance')

        # for each cluster id
        for cluster_id in range(1, max(clusters) + 1):
            self.topics.append([candidates[j] for j in range(len(clusters))
                                if clusters[j] == cluster_id])

        # assign cluster identifiers to candidates
        for i, cluster_id in enumerate(clusters):
            self.topic_identifiers[candidates[i]] = cluster_id - 1

    def calculate_bi_node_weights(self, node_i, node_j, weights):
        """

        :param node_i: str
        :param node_j: str
        :param weights: list
        :return:
        """
        for p_i in self.candidates[node_i].offsets:
            for p_j in self.candidates[node_j].offsets:
                # compute gap
                gap = abs(p_i - p_j)

                # alter gap according to candidate length
                target_node = node_i if p_i < p_j else node_j
                gap -= len(self.candidates[target_node].lexical_form) - 1

                weights.append(1.0 / gap)

    def build_topic_graph(self):
        """ Build the Multipartite graph. """

        # adding the nodes to the graph
        self.graph.add_nodes_from(self.candidates.keys())

        # pre-compute edge weights
        for node_i, node_j in combinations(self.candidates.keys(), 2):

            # discard intra-topic edges
            if self.topic_identifiers[node_i] == self.topic_identifiers[node_j]:
                continue

            weights = []
            self.calculate_bi_node_weights(node_i, node_j, weights)

            # add weighted edges
            if weights:
                # node_i -> node_j
                self.graph.add_edge(node_i, node_j, weight=sum(weights))
                # node_j -> node_i
                self.graph.add_edge(node_j, node_i, weight=sum(weights))

    def is_valid_edge(self, variant, first, end):
        return variant != first and self.graph.has_edge(variant, end)

    def find_connected_nodes(self, first, variants, weighted_edges):
        """find the nodes to which it connects

        :param first:
        :param variants:
        :param weighted_edges: dict
        :return:
        """
        for start, end in self.graph.edges(first):
            boosters = [self.graph[v][end]['weight'] for v in variants
                        if self.is_valid_edge(v, first, end)]

            if boosters:
                weighted_edges[(start, end)] = np.sum(boosters)

    def weight_adjustment(self, alpha=1.1):
        """ Adjust edge weights for boosting some candidates.

        :param alpha: float, hyper-parameter that controls the strength of the
            weight adjustment, defaults to 1.1.
        """
        weighted_edges = {}

        # Topical boosting
        for variants in self.topics:
            # skip one candidate topics
            if len(variants) == 1:
                continue

            # get the offsets
            offsets = [self.candidates[v].offsets[0] for v in variants]

            # get the first occurring variant
            first = variants[offsets.index(min(offsets))]

            # find the nodes to which it connects
            self.find_connected_nodes(first, variants, weighted_edges)

        # update edge weights
        for nodes, boosters in weighted_edges.items():
            node_i, node_j = nodes
            position_i = 1.0 / (1 + self.candidates[node_i].offsets[0])
            position_i = math.exp(position_i)
            self.graph[node_j][node_i]['weight'] += (boosters * alpha * position_i)

    def candidate_weighting(self,
                            threshold=0.74,
                            method='average',
                            alpha=1.1):
        """Candidate weight calculation using random walk.

        :param threshold: float, the minimum similarity for clustering, defaults to 0.25.
        :param method: str, the linkage method, defaults to average.
        :param alpha: float, hyper-parameter that controls the strength of the
            weight adjustment, defaults to 1.1.
        """
        if not self.candidates:
            return

        # cluster the candidates
        self.topic_clustering(threshold=threshold, method=method)

        # build the topic graph
        self.build_topic_graph()

        if alpha > 0.0:
            self.weight_adjustment(alpha)

        # compute the word scores using random walk
        self.weights = nx.pagerank_scipy(self.graph)

    def extract(self, input_file_or_string, n_best=10, pos=None, threshold=0.74):
        # 1. load document
        self.load_document(input=input_file_or_string, language='zh')

        # 2. clear previous graph and topics if exists
        self.clear_cache()

        # 3. select the longest sequences of nouns and adjectives, that do
        # not contain punctuation marks or stopwords as candidates.
        if pos is None:
            pos = {'n', 'a'}
        self.candidate_selection(pos=pos)

        # 4. build the Multipartite graph and rank candidates using random walk,
        # alpha controls the weight adjustment mechanism, see TopicRank for
        # threshold/method parameters.
        self.candidate_weighting(alpha=1.1,
                                 threshold=threshold,
                                 method='average')

        # 5. get the 10-highest scored candidates as keyphrases
        keyphrases = self.get_n_best(n=n_best)
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Word Rank module, main
"""
import codecs
import os
from loguru import logger
import re
from collections import Counter
from typing import Optional
from copy import deepcopy
import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report
from sklearn.model_selection import train_test_split

from pke_zh import USER_DATA_DIR
from pke_zh.utils.io_utils import load_pkl, save_pkl, save_json, load_json
from pke_zh.utils.text_utils import convert_to_unicode, is_number_string, is_alphabet_string, is_chinese_string
from pke_zh.utils.tokenizer import word_segment
from pke_zh.utils.file_utils import get_file
from pke_zh.unsupervised.tfidf import TfIdf
from pke_zh.unsupervised.textrank import TextRank

pwd_path = os.path.abspath(os.path.dirname(__file__))

# inner data file
default_stopwords_path = os.path.join(pwd_path, '../data/stopwords.txt')
person_name_path = os.path.join(pwd_path, '../data/person_name.txt')
place_name_path = os.path.join(pwd_path, '../data/place_name.txt')
common_char_path = os.path.join(pwd_path, '../data/common_char_set.txt')
pmi_path = os.path.join(pwd_path, '../data/pmi_word_score.json')
entropy_path = os.path.join(pwd_path, '../data/entropy_word_score.json')

# word rank model path
default_model_path = os.path.join(USER_DATA_DIR, 'wordrank_model.pkl')


class AttrDict(dict):
    """Dict that can get attribute by dot"""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class StatisticsFeature:
    """
    统计特征：包括PMI、IDF、textrank值、前后词互信息、左右邻熵、
    独立检索占比（term单独作为query的qv/所有包含term的query的qv和）、统计概率TF、idf变种iqf
    """

    def __init__(
            self,
            ngram=4,
            pmi_path=pmi_path,
            entropy_path=entropy_path,
            segment_sep=' ',
            stopwords_path=default_stopwords_path,
    ):
        self.tfidf_model = TfIdf(stopwords_path=stopwords_path)
        self.textrank_model = TextRank()
        self.segment_sep = segment_sep
        self.pmi_model = PMI(
            ngram=ngram,
            pmi_path=pmi_path,
            entropy_path=entropy_path
        )

    def _get_tags_score(self, word, tags):
        score = 0.0
        for i, s in tags:
            if word == i:
                score = s
        return score

    @staticmethod
    def read_text(file_path, col_sep=',', limit_len=1000000):
        text = ''
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                parts = line.split(col_sep)
                text += parts[0]
        if limit_len > 0:
            result = text[:limit_len]
        else:
            result = text
        return result

    def get_feature(self, sentence, is_word_segmented=False):
        """
        Get statistics feature
        :param sentence:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        sentence_features = {}
        rank_tags = self.textrank_model.extract(sentence)
        tfidf_tags = self.tfidf_model.extract(sentence)

        if is_word_segmented:
            word_seq = sentence.split(self.segment_sep)
        else:
            word_seq = word_segment(sentence, cut_type='word', pos=False)

        count = 0
        for word in word_seq:
            idf = self.tfidf_model.idf_freq.get(word, self.tfidf_model.median_idf)
            # PMI & Entropy
            left_word = word_seq[count - 1] if count > 0 else ''
            right_word = word_seq[count + 1] if count < len(word_seq) - 1 else ''
            left_right_word = ''.join([left_word, right_word])
            entropy_score = self.pmi_model.entropy_score(left_right_word)
            term_features.append(AttrDict(
                idf=idf,
                text_rank_score=self._get_tags_score(word, rank_tags),
                tfidf_score=self._get_tags_score(word, tfidf_tags),
                pmi_score=self.pmi_model.pmi_score(left_right_word),
                left_entropy_score=entropy_score[0],
                right_entropy_score=entropy_score[1],
            ))
            count += 1

        return term_features, sentence_features


class PMI:
    """
    前后词互信息PMI: 计算两个词语之间的关联程度，pmi越大越紧密
    左右邻熵：左右熵值越大，说明该词的周边词越丰富，意味着词的自由程度越大，其成为一个独立的词的可能性也就越大。
    """

    def __init__(
            self,
            text: str = None,
            ngram: int = 4,
            pmi_path: Optional[str] = pmi_path,
            entropy_path: Optional[str] = entropy_path
    ):
        if text is not None and len(text) > 0:
            logger.info(f'Use input text to generate new PMI dict, pmi path: {pmi_path}, entropy path: {entropy_path}')
            text = text.strip()
            words_freq = self.ngram_freq(text, ngram)
            self.pmi_score_dict = self.generate_pmi_score(words_freq, pmi_path)
            self.entropy_score_dict = self.generate_entropy_score(words_freq, text, entropy_path)
        else:
            self.pmi_score_dict = load_json(pmi_path)
            self.entropy_score_dict = load_json(entropy_path)
            logger.debug('Loaded PMI dict: %s' % pmi_path)

    @staticmethod
    def ngram_freq(text, ngram=4):
        """
        Get Ngram freq dict
        :param text: input text, sentence
        :param ngram: N
        :return: word frequency dict
        """
        words = []
        for i in range(1, ngram + 1):
            words += [text[j:j + i] for j in range(len(text) - i + 1)]
        words_freq = dict(Counter(words))
        return words_freq

    @staticmethod
    def generate_pmi_score(word_freq_dict, pmi_path=None):
        """
        Generate PMI score
        :param word_freq_dict: dict
        :return: dict(word, float), pmi score
        """
        result = dict()
        for word in word_freq_dict:
            if len(word) > 1:
                p_x_y = min([word_freq_dict.get(word[:i]) * word_freq_dict.get(word[i:]) for i in range(1, len(word))])
                score = p_x_y / word_freq_dict.get(word)
                result[word] = score
        if pmi_path:
            save_json(result, pmi_path)
            logger.info('Save pmi score to %s' % pmi_path)
        return result

    @staticmethod
    def entropy_list_score(char_list):
        """
        Get entropy score
        :param char_list: input char list
        :return: float, score
        """
        char_freq_dict = dict(Counter(char_list))
        char_size = len(char_list)
        entropy = -1 * sum([char_freq_dict.get(i) / char_size * np.log2(char_freq_dict.get(i) / char_size)
                            for i in char_freq_dict])
        return entropy

    def generate_entropy_score(self, word_freq_dict, text, entropy_path=None):
        """
        Generate entropy score
        :param word_freq_dict: dict
        :param text: input text, document
        :param entropy_path: save entropy file path
        :return: dict
        """
        result = dict()
        for word in word_freq_dict:
            if len(word) == 1:
                continue
            # pass error pattern
            if '*' in word:
                continue
            try:
                left_right_char = re.findall('(.)%s(.)' % word, text)
                left_char = [i[0] for i in left_right_char]
                left_entropy = self.entropy_list_score(left_char)

                right_char = [i[1] for i in left_right_char]
                right_entropy = self.entropy_list_score(right_char)
                if left_entropy > 0 or right_entropy > 0:
                    result[word] = [left_entropy, right_entropy]
            except Exception as e:
                logger.warning('error word %s, %s' % (word, e))
        if entropy_path:
            save_json(result, entropy_path)
            logger.info('Save entropy score to %s' % entropy_path)
        return result

    def pmi_score(self, word):
        """
        Get PMI score
        :param word:
        :return:
        """
        return self.pmi_score_dict.get(word, 0.0)

    def entropy_score(self, word):
        """
        Get entropy score
        :param word:
        :return:
        """
        return self.entropy_score_dict.get(word, [0.0, 0.0])


class TextFeature:
    """
    文本特征：包括Query长度、Term长度，Term在Query中的偏移量，term词性、长度信息、
    term数目、位置信息、句法依存tag、是否数字、是否英文、是否停用词、是否专名实体、
    是否重要行业词、embedding模长、删词差异度、以及短语生成树得到term权重等。
    """

    def __init__(
            self,
            stopwords_path=default_stopwords_path,
            person_name_path=person_name_path,
            place_name_path=place_name_path,
            common_char_path=common_char_path,
            segment_sep=' ',
    ):
        self.stopwords = self.load_set_file(stopwords_path)
        self.person_names = self.load_set_file(person_name_path)
        self.place_names = self.load_set_file(place_name_path)
        self.common_chars = self.load_set_file(common_char_path)
        self.segment_sep = segment_sep

    @staticmethod
    def load_set_file(path):
        words = set()
        with codecs.open(path, 'r', encoding='utf-8') as f:
            for w in f:
                w = w.strip()
                if w.startswith('#'):
                    continue
                if w:
                    words.add(w.split()[0])
        return words

    def is_stopword(self, word):
        return word in self.stopwords

    def is_name(self, word):
        names = self.person_names | self.place_names
        return word in names

    def is_entity(self, pos, entity_pos=('ns', 'n', 'vn', 'v')):
        return pos in entity_pos

    def is_common_char(self, c):
        return c in self.common_chars

    def is_common_char_string(self, word):
        return all(self.is_common_char(c) for c in word)

    def get_feature(self, query, is_word_segmented=False):
        """
        Get text feature
        :param query:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        if is_word_segmented:
            word_seq = query.split(self.segment_sep)
        else:
            word_seq = word_segment(query, cut_type='word', pos=False)
        # logger.debug('%s' % word_seq)

        # sentence
        sentence_features = AttrDict(
            query_length=len(query),
            term_size=len(word_seq),
        )

        # term
        idx = 0
        offset = 0
        for word in word_seq:
            word_list = deepcopy(word_seq)
            if word in word_list:
                word_list.remove(word)
            term_features.append(AttrDict(
                term=word,
                term_length=len(word),
                idx=idx,
                offset=offset,
                is_number=is_number_string(word),
                is_chinese=is_chinese_string(word),
                is_alphabet=is_alphabet_string(word),
                is_stopword=self.is_stopword(word),
                is_name=self.is_name(word),
                is_common_char=self.is_common_char_string(word),
            ))
            idx += len(word)
            offset += 1

        return term_features, sentence_features


class NGram:
    def __init__(self, model_name_or_path=None, cache_folder=os.path.expanduser('~/.pycorrector/datasets/')):
        if model_name_or_path and os.path.exists(model_name_or_path):
            logger.info('Load kenlm language model:{}'.format(model_name_or_path))
            language_model_path = model_name_or_path
        else:
            # 语言模型 2.95GB
            get_file(
                'zh_giga.no_cna_cmn.prune01244.klm',
                'https://deepspeech.bj.bcebos.com/zh_lm/zh_giga.no_cna_cmn.prune01244.klm',
                extract=True,
                cache_subdir=cache_folder,
                verbose=1)
            language_model_path = os.path.join(cache_folder, 'zh_giga.no_cna_cmn.prune01244.klm')
        try:
            import kenlm
        except ImportError:
            raise ImportError('Kenlm not installed, use "pip install -U kenlm".')
        self.lm = kenlm.Model(language_model_path)
        logger.debug('Loaded language model: %s.' % language_model_path)

    def ngram_score(self, sentence: str):
        """
        取n元文法得分
        :param sentence: str, 输入的句子
        :return:
        """
        return self.lm.score(' '.join(sentence), bos=False, eos=False)

    def perplexity(self, sentence: str):
        """
        取语言模型困惑度得分，越小句子越通顺
        :param sentence: str, 输入的句子
        :return:
        """
        return self.lm.perplexity(' '.join(sentence))


class LanguageFeature:
    """
    语言模型特征：整个query的语言模型概率/去掉该Term后的Query的语言模型概率。
    """

    def __init__(self, segment_sep=' '):
        self.segment_sep = segment_sep
        self.ngram = NGram()

    def get_feature(self, query, is_word_segmented=False):
        """
        Get language feature
        :param query:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        if is_word_segmented:
            word_seq = query.split(self.segment_sep)
        else:
            word_seq = word_segment(query, cut_type='word', pos=False)

        # sentence
        sentence_features = AttrDict(
            ppl=self.ngram.perplexity(word_seq),
        )

        # term
        count = 0
        for word in word_seq:
            word_list = deepcopy(word_seq)
            if word in word_list:
                word_list.remove(word)
            left_word = word_seq[count - 1] if count > 0 else ''
            right_word = word_seq[count + 1] if count < len(word_seq) - 1 else ''

            term_features.append(AttrDict(
                del_term_ppl=self.ngram.perplexity(word_list),
                term_ngram_score=self.ngram.ngram_score(word),
                left_term_score=self.ngram.ngram_score(left_word + word),
                right_term_score=self.ngram.ngram_score(word + right_word)
            ))
            count += 1
        return term_features, sentence_features


class WordRank:
    """WordRank keyphrase extraction model."""

    def __init__(
            self,
            model_path=default_model_path,
            stopwords_path=default_stopwords_path,
            person_name_path=person_name_path,
            place_name_path=place_name_path,
            common_char_path=common_char_path,
            segment_sep=' ',
            ngram=4,
            pmi_path=pmi_path,
            entropy_path=entropy_path,
    ):
        """ init word rank model

        Args:
            model_path (str): the path to load the model in pickle format,
                default to "~/.cache/pke_zh/wordrank_model.pkl".
        """
        self.text_feature = TextFeature(
            stopwords_path=stopwords_path,
            person_name_path=person_name_path,
            place_name_path=place_name_path,
            common_char_path=common_char_path,
            segment_sep=segment_sep
        )
        self.statistics_feature = StatisticsFeature(
            ngram=ngram,
            pmi_path=pmi_path,
            entropy_path=entropy_path,
            segment_sep=segment_sep,
            stopwords_path=stopwords_path,
        )
        self.language_feature = LanguageFeature()
        self.segment_sep = segment_sep
        self.model = None
        if not os.path.exists(model_path):
            # release 的模型
            cache_folder = os.path.abspath(os.path.dirname(model_path))
            file_name = 'wordrank_model.pkl'
            get_file(
                f'{file_name}',
                f'https://github.com/shibing624/pke_zh/releases/download/0.2.2/{file_name}',
                extract=False,
                cache_subdir=cache_folder,
                verbose=1)
            model_path = os.path.join(cache_folder, f'{file_name}')
        if model_path and os.path.exists(model_path):
            self.model = load_pkl(model_path)
            logger.debug('Loaded wordrank model: {}'.format(model_path))
        self.model_path = model_path

    @staticmethod
    def data_reader(file_path, col_sep=','):
        """
        Load data
        :param file_path:
        :param col_sep:
        :return: list, list: contents, labels
        """
        contents = []
        labels = []
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.rstrip()
                parts = line.split(col_sep)
                if len(parts) != 2:
                    continue
                contents.append(parts[0])
                labels.append(parts[1])
        return contents, labels

    def get_feature(self, query, is_word_segmented=False):
        """
        Get feature from query
        :param query: input query
        :param is_word_segmented: bool, is word segmented or not
        :return: features, terms
        """
        features = []
        terms = []

        text_terms, text_sent = self.text_feature.get_feature(query, is_word_segmented=is_word_segmented)
        stat_terms, stat_sent = self.statistics_feature.get_feature(query, is_word_segmented=is_word_segmented)
        lang_terms, lang_sent = self.language_feature.get_feature(query, is_word_segmented=is_word_segmented)
        # sentence feature
        text_sent.update(stat_sent)
        text_sent.update(lang_sent)
        # logger.debug('sentence features: %s' % text_sent)
        sent_feature = [text_sent.query_length, text_sent.term_size, text_sent.ppl]
        # term feature
        for text, stat, lang in zip(text_terms, stat_terms, lang_terms):
            text.update(stat)
            text.update(lang)
            # logger.debug('term features: %s' % text)
            term_feature = [
                text.term_length, text.idx, text.offset,
                float(text.is_number), float(text.is_chinese), float(text.is_alphabet),
                float(text.is_stopword), float(text.is_name), float(text.is_common_char),
                text.idf,
                text.text_rank_score, text.tfidf_score, text.pmi_score,
                text.left_entropy_score, text.right_entropy_score, text.del_term_ppl,
                text.term_ngram_score, text.left_term_score, text.right_term_score
            ]
            feature = sent_feature + term_feature
            features.append(feature)
            terms.append(text.term)
        return features, terms

    def train(self, train_file, col_sep=',', is_word_segmented=True):
        # 1.read train data
        contents, labels = self.data_reader(train_file, col_sep)
        logger.info('contents size:%s, labels size:%s' % (len(contents), len(labels)))

        features = []
        tags = []
        for content, label in zip(contents, labels):
            label_split = [int(i) for i in label.split(self.segment_sep)]
            content_split = content.split(self.segment_sep)
            if len(label_split) != len(content_split):
                logger.warning('pass, content size not equal label size, %s %s' % (content, label))
                continue
            tags += label_split
            content_features, terms = self.get_feature(content, is_word_segmented=is_word_segmented)
            features += content_features
        logger.info("[train]features size: %s, tags size: %s" % (len(features), len(tags)))
        assert len(features) == len(tags), "features size must equal tags size"
        X_train, X_val, y_train, y_val = train_test_split(features, tags, test_size=0.1, random_state=0)
        logger.debug("train size:%s, val size:%s" % (len(y_train), len(y_val)))
        # 3.train classification model, save model file
        model = RandomForestClassifier(n_estimators=300)
        # fit
        logger.debug("start train model ...")
        model.fit(X_train, y_train)
        # save model
        save_pkl(model, self.model_path, overwrite=True)
        logger.info("model saved: %s" % self.model_path)

        # 4.validation and evaluate
        logger.debug("evaluate model with validation data")
        self.evaluate(model, X_val, y_val)
        self.model = model
        return model

    def predict(self, query, is_word_segmented=False):
        logger.info('model predict')
        features, terms = self.get_feature(query, is_word_segmented=is_word_segmented)
        # predict classification model
        if self.model_path and os.path.exists(self.model_path):
            self.model = load_pkl(self.model_path)
            logger.debug('Loaded model: {}'.format(self.model_path))
        else:
            logger.error('model not found. path: {}'.format(self.model_path))
        logger.debug("model predict")
        label_pred = self.model.predict(features)
        logger.info("words: %s" % terms)
        logger.info("predict label: %s" % label_pred)
        return label_pred

    @staticmethod
    def evaluate(model, test_data, test_label, pred_save_path=None):
        print('{0}, val mean acc:{1}'.format(model.__str__(), model.score(test_data, test_label)))
        # multi
        label_pred = model.predict(test_data)
        # precision_recall_curve: multiclass format is not supported
        print(classification_report(test_label, label_pred))
        if pred_save_path:
            with open(pred_save_path, 'w', encoding='utf-8') as f:
                for i in label_pred:
                    f.write(str(i) + '\n')
        return label_pred

    def extract(self, input_string, n_best=10):
        """
        Extract keywords from text
        :param input_string:
        :param n_best:
        :return:
        """
        term_weights = []
        text = convert_to_unicode(input_string)
        if len(text) == 1:
            term_scores = zip([text], [0.0])
        else:
            # get feature
            data_feature, terms = self.get_feature(text, is_word_segmented=False)
            # predict model
            label_pred = self.model.predict(data_feature)
            term_scores = zip(terms, label_pred)
        for w, p in term_scores:
            if w not in term_weights:
                term_weights.append((w, p))

        keyphrases = sorted(term_weights, key=lambda k: k[1], reverse=True)[:n_best]
        return keyphrases



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Data structures
modify from: https://github.com/boudinfl/pke/blob/master/pke/data_structures.py
"""


class Sentence(object):
    """The sentence data structure."""

    def __init__(self, words):

        self.words = words
        """list of words (tokens) in the sentence."""

        self.pos = []
        """list of Part-Of-Speeches."""

        self.length = len(words)
        """length (number of tokens) of the sentence."""

        self.meta = {}
        """meta-information of the sentence."""

    @staticmethod
    def _is_length_word_pos_meta_not_equal(length1, words1, pos1, meta1, length2, words2, pos2, meta2):
        if length1 != length2 or \
                words1 != words2 or \
                pos1 != pos2 or \
                meta1 != meta2:
            return True
        return False

    def __eq__(self, other):
        """Compares two sentences for equality."""

        # 1. test whether they are instances of different classes
        # 2. test whether they are of same length
        # 3. test whether they have the same words
        # 4. test whether they have the same PoS tags
        # 5. test whether they have the same meta-information
        if type(self) != type(other) or \
                self._is_length_word_pos_meta_not_equal(
                    self.length, self.words, self.pos, self.meta,
                    other.length, other.words, other.pos, other.meta
                ):
            # self.length != other.length or \
            # self.words != other.words or \
            # self.pos != other.pos or \
            # self.meta != other.meta:
            return False

        # if everything is ok then they are equal
        return True


class Candidate(object):
    """The keyphrase candidate data structure."""

    def __init__(self):
        self.surface_forms = []
        """ the surface forms of the candidate. """

        self.offsets = []
        """ the offsets of the surface forms. """

        self.sentence_ids = []
        """ the sentence id of each surface form. """

        self.pos_patterns = []
        """ the Part-Of-Speech patterns of the candidate. """

        self.lexical_form = []
        """ the lexical form of the candidate. """


def parse_sentence(sent_obj):
    s = Sentence(words=sent_obj['words'])

    # add the POS
    s.pos = sent_obj['POS']

    # add the meta-information, for chinese, it's `char_offset`
    m_key = "char_offsets"
    s.meta[m_key] = sent_obj[m_key]

    return s


class Document(object):
    """The Document data structure."""

    def __init__(self):

        self.input_file = None
        """ The path of the input file. """

        self.sentences = []
        """ The sentence container (list of Sentence). """

    @staticmethod
    def from_sentences(sentences, **kwargs):
        """Populate the sentence list.

        Args:
            sentences (Sentence list): content to create the document.
            input_file (str): path to the input file.
        """

        # initialize document
        doc = Document()

        # set the input file
        doc.input_file = kwargs.get('input_file', None)

        # loop the parsed sentences
        doc.sentences = [parse_sentence(sent) for sent in sentences]

        return doc

    def __eq__(self, other):
        """Compares two documents for equality."""

        # test whether they are instances of different classes
        if type(self) != type(other):
            return False

        # test whether they have the same input path
        if self.input_file != other.input_file:
            return False

        # test whether they contain the same lists of sentences
        if self.sentences != other.sentences:
            return False

        # if everything is ok then they are equal
        return True
